#include "WeissHapticSensor.h"
#include "TransmissionException.h"
#include <boost/regex.hpp>
#include <boost/format.hpp>

using namespace armarx;

WeissHapticSensor::WeissHapticSensor(std::string device)
    : device(device), connected(false)
{
    sensorTask = new RunningTask<WeissHapticSensor>(this, &WeissHapticSensor::frameAcquisitionTaskLoop);
    boost::smatch match;
    boost::regex_search( device, match, boost::regex("\\w+$") );
    this->deviceFileName = match[0];
}

void WeissHapticSensor::connect()
{

    //cout << "Open Serial" << endl;
    this->interface.reset(new SerialInterface(device.c_str(), 115200));
    interface->startLogging(deviceFileName + ".transmission.log");
    interface->open();
    //cout << *interface << endl;
    this->sensor.reset(new TactileSensor(interface));

    jsWriter.reset(new TextWriter(deviceFileName + ".js"));

    //cout << "Stop Periodic Frame Acquisition" << endl;
    sensor->stopPeriodicFrameAcquisition();

    //string sensorType = sensor->getSensorType();
    //cout << boost::format("Sensor Type = %1%") % sensorType << endl;

    //tac_system_information_t si = sensor->getSystemInformation();
    //TactileSensor::printSystemInformation(si);

    //sensor->setDeviceTag("Test Tag");
    this->tag = sensor->getDeviceTag();
    //cout << boost::format("Sensor Tag = %1%") % tag << endl;

    jsWriter->writeLine(str(boost::format("%s = {};") % deviceFileName));
    jsWriter->writeLine(str(boost::format("sensors.push(%s);") % deviceFileName));
    jsWriter->writeLine(str(boost::format("%s.device = \"%s\";") % deviceFileName % device));
    jsWriter->writeLine(str(boost::format("%s.tag = \"%s\";") % deviceFileName % tag));
    jsWriter->writeLine(str(boost::format("%s.data = [];") % deviceFileName));

    //cout << "Tare Sensor Matrix" << endl;
    //sensor->tareSensorMatrix(1);

    //cout << "Get Matrix Information" << endl;
    this->mi = sensor->getMatrixInformation();
    TactileSensor::printMatrixInfo(&mi);

    sensor->setAquisitionWindow(1, 1, mi.res_x, mi.res_y);

    sensor->setFrontEndGain(255);

    ARMARX_LOG << "Front end gain set to " << (int)sensor->getFrontEndGain();

    connected = true;
    cout << this << ": Connect done, Interface=" << sensor->getInterfaceInfo() << endl;
}

void WeissHapticSensor::disconnect()
{
    connected = false;
    sensorTask->stop(true);
}

void WeissHapticSensor::setListenerPrx(HapticUnitListenerPrx listenerPrx)
{
    this->listenerPrx = listenerPrx;
}

void WeissHapticSensor::startSampling()
{
    cout << this << ": startSampling" << endl;
    sensorTask->start();
}

void WeissHapticSensor::frameAcquisitionTaskLoop()
{
    cout << this << ": readAndReportSensorValues" << endl;
    //bool periodic = false;
    cout << sensor->getInterfaceInfo() << endl;

    cout << this << ": startPeriodicFrameAcquisition" << endl;
    sensor->startPeriodicFrameAcquisition(0);

    while(!sensorTask->isStopped())
    {
        //ARMARX_INFO << deactivateSpam(1) << this << ": receicePeriodicFrame";

        try
        {
            //long start = TimestampVariant::nowLong();
            PeriodicFrameData data = sensor->receicePeriodicFrame();
            //long end = TimestampVariant::nowLong();
            //cout << end - start << endl;

            MatrixFloatPtr matrix = new MatrixFloat(mi.res_y, mi.res_x);
            for (int y = 0; y < mi.res_y; y++)
            {
                for (int x = 0; x < mi.res_x; x++)
                {
                    short val = (*data.data)[y * mi.res_x + x];
                    (*matrix)(y, x) = val;
                }
            }

            TimestampVariantPtr now = TimestampVariant::nowPtr();
            writeMatrixToJs(matrix, now);
            listenerPrx->reportSensorValues(device, tag, matrix, now);
        }
        catch(ChecksumErrorException)
        {
            ARMARX_WARNING << "Caught ChecksumErrorException on " << device << ", skipping frame";
        }
        //usleep(3000);

        //usleep(3000);
        //usleep(1000000);
    }

    cout << this << ": stopPeriodicFrameAcquisition" << endl;
    sensor->stopPeriodicFrameAcquisition();
}

void WeissHapticSensor::writeMatrixToJs(MatrixFloatPtr matrix, TimestampVariantPtr timestamp)
{
    //std::cout << "writeMatrixToJs" << std::endl;
    if(jsWriter != NULL)
    {
        jsWriter->writeLine(str(boost::format("%s.data.push([%i, %s]);") % deviceFileName % timestamp->getTimestamp() % matrix->toJsonRowMajor()));
    }
}
