/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotStateComponent::
 * @author     ( at kit dot edu)
 * @date
 * @copyright  http://www.gnu.org/licenses/gpl.txt
 *             GNU General Public License
 */

#include "RobotStateComponent.h"
#include <boost/foreach.hpp>
#include <VirtualRobot/XML/RobotIO.h>
#include <VirtualRobot/Nodes/RobotNode.h>
#include <VirtualRobot/RobotNodeSet.h>
#include <boost/format.hpp>
#include <boost/foreach.hpp>
#include <Core/core/system/ArmarXDataPath.h>

using namespace std;
using namespace VirtualRobot;
using namespace Eigen;
using namespace Ice;
using namespace boost;

std::ofstream out_file;

namespace armarx
{
    RobotStateComponent::~RobotStateComponent()
    {
        if (_synchronizedPrx)
        {
            _synchronizedPrx->unref();
        }
        ARMARX_VERBOSE << "Destructor";
    }


    void RobotStateComponent::onInitComponent()
    {
        std::string robotFile = getProperty<std::string>("RobotFileName").getValue();
        if (!ArmarXDataPath::getAbsolutePath(robotFile,robotFile))
        {
            throw UserException("Could not find robot file " + robotFile);
        }

        this->_synchronized = VirtualRobot::RobotIO::loadRobot(robotFile, VirtualRobot::RobotIO::eStructure);

        if (this->_synchronized)
        {
            ARMARX_VERBOSE << "Loaded robot from file " << robotFile;
        }
        else
        {
            ARMARX_VERBOSE << "Failed loading robot from file " << robotFile;
        }

        string robotNodeSetName = getProperty<std::string>("RobotNodeSetName").getValue();


        if(robotNodeSetName == "")
        {
            throw UserException("RobotNodeSet not defined");
        }

        VirtualRobot::RobotNodeSetPtr rns =  this->_synchronized->getRobotNodeSet(robotNodeSetName);
        if (!rns)
        {
             throw UserException("RobotNodeSet not defined");
        }

        vector<RobotNodePtr> nodes = rns->getAllRobotNodes();

        ARMARX_VERBOSE << "Using RobotNodeSet " << robotNodeSetName << " with " << nodes.size() << " robot nodes.";
        BOOST_FOREACH(RobotNodePtr node, nodes)
        {
            ARMARX_VERBOSE << "Node: " << node->getName() << endl;
        }
        usingTopic(robotNodeSetName + "State");

        VirtualRobot::RobotNodeSetPtr pns = this->_synchronized->getRobotNodeSet("Platform");
        if (pns)
        {
            usingTopic("PlatformState");
            vector<RobotNodePtr> nodes = pns->getAllRobotNodes();

            ARMARX_VERBOSE << "Using RobotNodeSet Platform with " << nodes.size() << " robot nodes.";
            BOOST_FOREACH(RobotNodePtr node, nodes)
            {
                ARMARX_VERBOSE << "Node: " << node->getName() << endl;
            }
        }
    }


    void RobotStateComponent::onConnectComponent()
    {
        ARMARX_INFO << "Started RobotStateComponent" << flush;
    }


    SharedRobotInterfacePrx RobotStateComponent::getSynchronizedRobot(const Current & current)
    {
        if (!this->_synchronizedPrx)
        {
            SharedRobotInterfacePtr shared =  new SharedRobotServant(this->_synchronized);
            shared->ref();
            this->_synchronizedPrx = SharedRobotInterfacePrx::uncheckedCast(current.adapter->addWithUUID(shared));
        }
        return this->_synchronizedPrx;
    }


    SharedRobotInterfacePrx RobotStateComponent::getRobotSnapshot(const string & time,const Current & current)
    {
        SharedRobotInterfacePtr p = new SharedRobotServant(this->_synchronized->clone(time));
        return SharedRobotInterfacePrx::uncheckedCast(current.adapter->addWithUUID(p));
    }

    void RobotStateComponent::reportJointAngles(const NameValueMap& jointAngles, bool aValueChanged, const Current& c)
    {
//        IceUtil::Time start = IceUtil::Time::now();
        if(aValueChanged)
        {
            ReadLockPtr lock = _synchronized->getReadLock();

            std::vector<float> jv;
            std::vector< RobotNodePtr > nodes;
            BOOST_FOREACH(NameValueMap::value_type namedAngle, jointAngles )
            {
                RobotNodePtr node = this->_synchronized->getRobotNode(namedAngle.first);
                nodes.push_back(node);
                jv.push_back(namedAngle.second);
            }
            _synchronized->setJointValues(nodes, jv);
        }
//        ARMARX_VERBOSE_S << "duration: " << (IceUtil::Time::now() - start).toMilliSeconds() << " ms";
    }


    void RobotStateComponent::reportControlModeChanged(const NameControlModeMap& jointModes,  bool aValueChanged,const Current& c){}
    void RobotStateComponent::reportJointVelocities(const NameValueMap& jointVelocities,  bool aValueChanged,const Current& c){}
    void RobotStateComponent::reportJointTorques(const NameValueMap& jointTorques,  bool aValueChanged,const Current& c){}
    void RobotStateComponent::reportJointCurrents(const NameValueMap& jointCurrents,  bool aValueChanged,const Current& c){}
    void RobotStateComponent::reportJointMotorTemperatures(const NameValueMap& jointMotorTemperatures, bool aValueChanged, const Current& c){}
    void RobotStateComponent::reportJointStatuses(const NameStatusMap& jointStatuses, bool aValueChanged, const Current& c){}

    void RobotStateComponent::reportPlatformPose(float currentPlatformPositionX, float currentPlatformPositionY, float currentPlatformRotation, const Ice::Current& c)
    {
        if (!this->_synchronized->hasRobotNodeSet("Platform") && !this->_synchronized->hasRobotNode("Platform"))
        {
            return;
        }
        {
            WriteLockPtr lock = _synchronized->getWriteLock();
            _synchronized->setJointValue("X_Platform", currentPlatformPositionX);
            _synchronized->setJointValue("Y_Platform", currentPlatformPositionY);
            _synchronized->setJointValue("Yaw_Platform", currentPlatformRotation);
            _synchronized->applyJointValues();
        }
    }

    void RobotStateComponent::reportNewTargetPose(float newPlatformPositionX, float newPlatformPositionY, float newPlatformRotation, const Ice::Current& c)
    {
    }

    void RobotStateComponent::reportPlatformVelocity(float currentPlatformVelocityX, float currentPlatformVelocityY, float currentPlatformVelocityRotation, const Ice::Current &c)
    {

    }

    PropertyDefinitionsPtr RobotStateComponent::createPropertyDefinitions()
    {
        return PropertyDefinitionsPtr(new RobotStatePropertyDefinitions(
                                               getConfigIdentifier()));
    }
}
