#ifndef __SHARED_ROBOT_SERVANTS_H__
#define __SHARED_ROBOT_SERVANTS_H__

#include <Core/core/Component.h>
#include <Core/core/system/ImportExportComponent.h>
#include <RobotAPI/libraries/robotstate/remote/ArmarPose.h>


#include <RobotAPI/interface/robotstate/RobotState.h>

#include <boost/thread.hpp>

namespace VirtualRobot
{
    class Robot;
    typedef boost::shared_ptr<Robot> RobotPtr;
    class RobotNode;
    typedef boost::shared_ptr<RobotNode> RobotNodePtr;
}

namespace armarx {
    // Zugriff muss mutex geschuetzt werden!

    /**
     * @brief The base class for remote objects on the ice server
     *
     */

    class SharedObjectBase :
        virtual public SharedObjectInterface
    {
        public:
            virtual void ref(const Ice::Current &c);
            virtual void unref(const Ice::Current &c);
            virtual void destroy(const Ice::Current &c);
            SharedObjectBase();
        private:
            unsigned int _referenceCount;
            boost::mutex _counterMutex;
    };

    /**
     * @brief The SharedRobotNodeServant class is a remote represenation of a Simox VirtualRobot::Robot
     *
     * @details This class is used only internally by the the RobotStateComponent. Other classes such as the LinkedPose, RemoteRobot,
     * TCPControlUnit and HeadIKUnit classes address this class by the SharedRobotNodeInterface and SharedRobotNodeInterfacePrx generated by
     * ICE.
     */
    class SharedRobotNodeServant :
        virtual public SharedRobotNodeInterface,
        public SharedObjectBase
    {
    public:
        SharedRobotNodeServant(VirtualRobot::RobotNodePtr node /*,const Ice::Current &current*/);
        ~SharedRobotNodeServant();

        virtual float getJointValue(const Ice::Current &current) const;
        virtual std::string getName(const Ice::Current &current) const;

        virtual PoseBasePtr getLocalTransformation(const Ice::Current &current) const;
        //virtual PoseBasePtr getPostJointTransformation(const Ice::Current &current) const;
        virtual PoseBasePtr getGlobalPose(const Ice::Current &current) const;
        virtual PoseBasePtr getGlobalPoseJoint(const Ice::Current &current) const;

        virtual JointType getType(const Ice::Current &current) const;
        virtual Vector3BasePtr getJointTranslationDirection(const Ice::Current &current) const;
        virtual Vector3BasePtr getJointRotationAxis(const Ice::Current &current) const;

        virtual bool hasChild(const std::string &name, bool recursive, const Ice::Current &current) const;
        virtual std::string getParent(const Ice::Current &current) const;
        virtual NameList getChildren(const Ice::Current &current) const;
        virtual NameList getAllParents(const std::string &name, const Ice::Current &current) const;

        virtual float getJointValueOffest(const Ice::Current &current) const;
        virtual float getJointLimitHigh(const Ice::Current &current) const;
        virtual float getJointLimitLow(const Ice::Current &current) const;



    protected:
        VirtualRobot::RobotNodePtr _node;
    };

    /**
     * @brief The SharedRobotNodeServant class is a remote represenation of a Simox VirtualRobot::Robot
     *
     * @details This class is used only internally by the the RobotStateComponent. The RemoteRobot class SharedRobotInterface and SharedRobotInterfacePrx
     * classes generated by Ice.
     */

    class   SharedRobotServant :
        public virtual SharedRobotInterface,
        public SharedObjectBase
    {
    public:
        SharedRobotServant(VirtualRobot::RobotPtr robot);
        ~SharedRobotServant();

        virtual SharedRobotNodeInterfacePrx getRobotNode(const std::string & name, const Ice::Current &current);
        virtual SharedRobotNodeInterfacePrx getRootNode(const Ice::Current &current);
        virtual bool hasRobotNode(const std::string & name, const Ice::Current &current);

        virtual NameList getRobotNodes(const Ice::Current &current);
        virtual RobotNodeSetInfoPtr getRobotNodeSet(const std::string & name, const Ice::Current &current);
        virtual NameList getRobotNodeSets(const Ice::Current &current);
        virtual bool hasRobotNodeSet(const std::string & name, const Ice::Current &current);

        virtual std::string getName(const Ice::Current &current);
        virtual std::string getType(const Ice::Current &current);

        virtual PoseBasePtr getGlobalPose(const Ice::Current &current);
        NameValueMap getConfig(const Ice::Current &current);

        VirtualRobot::RobotPtr getRobot()
        {
            return this->_robot;
        }

    protected:
        VirtualRobot::RobotPtr _robot;
        boost::recursive_mutex m;
        std::map<std::string, SharedRobotNodeInterfacePrx> _cachedNodes;
    };
}

#endif
