#include "ArmarPose.h"

// boost includes
#include <boost/property_tree/ptree.hpp>
#include <boost/property_tree/xml_parser.hpp>



#include <VirtualRobot/Robot.h>
#include <VirtualRobot/LinkedCoordinate.h>
#include <VirtualRobot/LinkedCoordinate.h>
#include <VirtualRobot/VirtualRobot.h>

using namespace Eigen;
using namespace std;

namespace armarx
{

    Vector3::Vector3()
    {
        x = 0;
        y = 0;
        z = 0;
    }

    Vector3::Vector3(const Vector3f & v)
    {
        x = v(0);
        y = v(1);
        z = v(2);
    }

    Vector3::Vector3(const Matrix4f & m)
    {
        x = m(0,3);
        y = m(1,3);
        z = m(2,3);
    }

    Vector3f Vector3::toEigen() const
    {
        Vector3f v;
        v << this->x,this->y,this->z;
        return v;
    }

    string Vector3::output(const Ice::Current &c) const
    {
        std::stringstream s;
        s << toEigen();
        return s.str();
    }

    int Vector3::readFromXML(const string &xmlData, const Ice::Current &c)
    {
        boost::property_tree::ptree pt = Variant::GetPropertyTree(xmlData);

        x = (pt.get<float>("x"));
        y = (pt.get<float>("y"));
        z = (pt.get<float>("z"));
        return 1;
    }

    string Vector3::writeAsXML(const Ice::Current &)
    {
        using namespace boost::property_tree;
        ptree pt;
        pt.add("x", x);
        pt.add("y", y);
        pt.add("z", z);

        std::stringstream stream;
        xml_parser::write_xml(stream, pt);
        return stream.str();
    }

    void Vector3::serialize(const ObjectSerializerBasePtr& serializer, const ::Ice::Current&) const
    {
        AbstractObjectSerializerPtr obj = AbstractObjectSerializerPtr::dynamicCast(serializer);

        obj->setFloat("x", x);
        obj->setFloat("y", y);
        obj->setFloat("z", z);
    }

    void Vector3::deserialize(const ObjectSerializerBasePtr& serializer, const ::Ice::Current& c)
    {
        AbstractObjectSerializerPtr obj = AbstractObjectSerializerPtr::dynamicCast(serializer);

        x = obj->getFloat("x");
        y = obj->getFloat("y");
        z = obj->getFloat("z");
    }


    Quaternion::Quaternion()
    {
    }

    Quaternion::Quaternion(const Matrix4f &m4)
    {
        Matrix3f m3=m4.block<3,3>(0,0);
        this->init(m3);
    }

    Quaternion::Quaternion(const Matrix3f &m3)
    {
        this->init(m3);
    }

    Quaternion::Quaternion(const Eigen::Quaternionf &q)
    {
        this->init(q);
    }

    Matrix3f Quaternion::toEigen() const
    {
        Matrix3f rot;
        rot = Quaternionf(
                    this->qw,
                    this->qx,
                    this->qy,
                    this->qz);
        return rot;
    }

    void Quaternion::init(const Matrix3f &m3)
    {
        Quaternionf q;
        q = m3;
        init(q);
    }

    void Quaternion::init(const Eigen::Quaternionf &q)
    {
        this->qw=q.w();
        this->qx=q.x();
        this->qy=q.y();
        this->qz=q.z();
    }

    Matrix3f Quaternion::slerp(float alpha, const Eigen::Matrix3f &m)
    {
        return Quaternion::slerp(alpha,this->toEigen(),m);
    }

    Matrix3f Quaternion::slerp(float alpha, const Eigen::Matrix3f &m1, const Eigen::Matrix3f &m2)
    {
        Matrix3f result;
        Quaternionf q1,q2;
        q1 = m1;
        q2=m2;
        result= q1.slerp(alpha,q2);
        return result;
    }

    string Quaternion::output(const Ice::Current &c) const
    {
        std::stringstream s;
        s << toEigen();
        return s.str();
    }

    int Quaternion::readFromXML(const string &xmlData, const Ice::Current &c)
    {
        boost::property_tree::ptree pt = Variant::GetPropertyTree(xmlData);

        using namespace Eigen;

        Quaternionf q;
        if(pt.get_optional<float>("angle"))
        {// AxisAngle-Notation
            float angle = pt.get<float>("angle");

            Vector3f axis;
            axis << pt.get<float>("x")  , pt.get<float>("y"), pt.get<float>("z");

            AngleAxisf aa(angle, axis);
            q = aa;
            qw = q.w();
            qx = q.x();
            qy = q.y();
            qz = q.z();
        }
        else
        {
            qx = (pt.get<float>("qx"));
            qy = (pt.get<float>("qy"));
            qz = (pt.get<float>("qz"));
            qw = (pt.get<float>("qw"));
        }

        return 1;
    }

    string Quaternion::writeAsXML(const Ice::Current &)
    {
        using namespace boost::property_tree;
        ptree pt;
        pt.add("qx", qx);
        pt.add("qy", qy);
        pt.add("qz", qz);
        pt.add("qw", qw);

        std::stringstream stream;
        xml_parser::write_xml(stream, pt);
        return stream.str();
    }

    void Quaternion::serialize(const ObjectSerializerBasePtr& serializer, const ::Ice::Current&) const
    {
        AbstractObjectSerializerPtr obj = AbstractObjectSerializerPtr::dynamicCast(serializer);

        obj->setFloat("qx", qx);
        obj->setFloat("qy", qy);
        obj->setFloat("qz", qz);
        obj->setFloat("qw", qw);
    }

    void Quaternion::deserialize(const ObjectSerializerBasePtr& serializer, const ::Ice::Current& c)
    {
        AbstractObjectSerializerPtr obj = AbstractObjectSerializerPtr::dynamicCast(serializer);

        qx = obj->getFloat("qx");
        qy = obj->getFloat("qy");
        qz = obj->getFloat("qz");
        qw = obj->getFloat("qw");
    }

    Pose::Pose(const Eigen::Matrix3f &m, const Eigen::Vector3f &v)
    {
        this->orientation = new Quaternion(m);
        this->position = new Vector3(v);

        //The initializiation has to take place ALSO after the demarshalling
        this->c_position = Vector3Ptr::dynamicCast(this->position);
        this->c_orientation = QuaternionPtr::dynamicCast(this->orientation);
    }

    Pose::Pose()
    {
        this->orientation = new Quaternion();
        this->position = new Vector3();
        this->c_position = Vector3Ptr::dynamicCast(this->position);
        this->c_orientation = QuaternionPtr::dynamicCast(this->orientation);
    }

    Pose::Pose(const Matrix4f &m)
    {
        this->orientation = new Quaternion(m);
        this->position = new Vector3(m);
        this->c_position = Vector3Ptr::dynamicCast(this->position);
        this->c_orientation = QuaternionPtr::dynamicCast(this->orientation);
    }

    Matrix4f Pose::toEigen() const
    {
        Matrix4f m = Matrix4f::Identity();
        // WARNING dynamicCast is slow! To speed this up (if necessary)
        // create Member variables in the demarshalization.
        m.block<3,3>(0,0) = QuaternionPtr::dynamicCast(this->orientation)->toEigen();
        m.block<3,1>(0,3) = Vector3Ptr::dynamicCast(this->position)->toEigen();
        return m;
    }

    string Pose::output(const Ice::Current &c) const
    {
        std::stringstream s;
        s << toEigen();
        return s.str();
    }

    int Pose::readFromXML(const string &xmlData, const Ice::Current &c)
    {
        position->readFromXML(xmlData);
        orientation->readFromXML(xmlData);

        return 1;
    }

    string Pose::writeAsXML(const Ice::Current &)
    {
        using namespace boost::property_tree;
        std::stringstream stream;
        stream <<
                  position->writeAsXML() <<
                  orientation->writeAsXML();
        return stream.str();
    }

    void Pose::serialize(const ObjectSerializerBasePtr& serializer, const ::Ice::Current& c) const
    {
        position->serialize(serializer, c);
        orientation->serialize(serializer, c);
    }

    void Pose::deserialize(const ObjectSerializerBasePtr& serializer, const ::Ice::Current& c)
    {
        AbstractObjectSerializerPtr obj = AbstractObjectSerializerPtr::dynamicCast(serializer);

        position->deserialize(obj);
        orientation->deserialize(obj);
    }

    FramedVector3::FramedVector3()
    {
    }

    FramedVector3::FramedVector3(const FramedVector3 &source) :
        IceUtil::Shared(source),
        Vector3Base(source),
        FramedVector3Base(source),
        Vector3(source)
    {
    }

    FramedVector3::FramedVector3(const Eigen::Vector3f &vec, const string &frame) :
        Vector3(vec)
    {
        this->frame = frame;
    }

    string FramedVector3::getFrame() const
    {
        return frame;
    }

    FramedVector3Ptr FramedVector3::ChangeFrame(const VirtualRobot::RobotPtr robot, const FramedVector3 &framedVec, const string &newFrame)
    {
        if(framedVec.getFrame() == newFrame)
            return FramedVector3Ptr::dynamicCast(framedVec.clone());

        Eigen::Matrix4f rotToNewFrame = __GetRotationBetweenFrames(framedVec.frame, newFrame, robot);

        Eigen::Vector3f vecOldFrame = framedVec.Vector3::toEigen();

        Eigen::Vector3f newVec = rotToNewFrame.block(0,0,3,3).inverse() * vecOldFrame;

        FramedVector3Ptr result = new FramedVector3();
        result->x = newVec(0);
        result->y = newVec(1);
        result->z = newVec(2);
        result->frame = newFrame;
        return result;
    }

    string FramedVector3::output(const Ice::Current &c) const
    {
        std::stringstream s;
        s << toEigen() << std::endl << "frame: " << getFrame();
        return s.str();
    }

    Eigen::Matrix4f FramedVector3::__GetRotationBetweenFrames(const std::string &oldFrame, const std::string &newFrame, VirtualRobot::RobotPtr robotState)
    {
        Eigen::Matrix4f toNewFrame = robotState->getRobotNode(oldFrame)->getTransformationTo(robotState->getRobotNode(newFrame));
        toNewFrame(0,3) = 0;
        toNewFrame(1,3) = 0;
        toNewFrame(2,3) = 0;

        return toNewFrame;
    }

    int FramedVector3::readFromXML(const string &xmlData, const Ice::Current &c)
    {
        int result = Vector3::readFromXML(xmlData);

        boost::property_tree::ptree pt = Variant::GetPropertyTree(xmlData);

        frame = (pt.get<std::string>("frame"));

        return result;
    }

    string FramedVector3::writeAsXML(const Ice::Current &c)
    {
        using namespace boost::property_tree;
        ptree pt = Variant::GetPropertyTree(Vector3::writeAsXML());
        pt.add("frame", frame);

        std::stringstream stream;
        boost::property_tree::xml_parser::xml_writer_settings<char> settings('\t', 1);
        xml_parser::write_xml(stream, pt, settings);

        return stream.str();
    }

    void FramedVector3::serialize(const ObjectSerializerBasePtr &serializer, const Ice::Current &) const
    {
        AbstractObjectSerializerPtr obj = AbstractObjectSerializerPtr::dynamicCast(serializer);
        Vector3::serialize(serializer);
        obj->setString("frame", frame);

    }

    void FramedVector3::deserialize(const ObjectSerializerBasePtr &serializer, const Ice::Current &)
    {
        AbstractObjectSerializerPtr obj = AbstractObjectSerializerPtr::dynamicCast(serializer);
        Vector3::deserialize(serializer);
        frame = obj->getString("frame");
    }



    FramedPose::FramedPose() :
        Pose()
    {
        frame = "";
    }

    FramedPose::FramedPose(const Eigen::Matrix3f& m, const Eigen::Vector3f& v, const std::string& s) :
        Pose(m, v)
    {
        frame = s;
    }

    FramedPose::FramedPose(const Eigen::Matrix4f& m, const std::string& s) :
        Pose(m)
    {
        frame = s;
    }

    std::string FramedPose::getFrame() const
    {
        return frame;
    }

    string FramedPose::output(const Ice::Current &c) const
    {
        std::stringstream s;
        s << toEigen() << std::endl << "frame: " << getFrame();
        return s.str();
    }

    int FramedPose::readFromXML(const string &xmlData, const Ice::Current &c){
        boost::property_tree::ptree pt = Variant::GetPropertyTree(xmlData);


        Pose::readFromXML(xmlData);
        frame = (pt.get<std::string>("frame"));

        return 1;
    }

    string FramedPose::writeAsXML(const Ice::Current &)
    {
        using namespace boost::property_tree;
        std::stringstream stream;
        stream << Pose::writeAsXML()
               << "<frame>"<<frame<<"</frame>";
        return stream.str();
    }

    void FramedPose::serialize(const ObjectSerializerBasePtr& serializer, const ::Ice::Current& c) const
    {
        AbstractObjectSerializerPtr obj = AbstractObjectSerializerPtr::dynamicCast(serializer);

        Pose::serialize(obj, c);
        obj->setString("frame", frame);
    }

    void FramedPose::deserialize(const ObjectSerializerBasePtr& serializer, const ::Ice::Current& c)
    {
        AbstractObjectSerializerPtr obj = AbstractObjectSerializerPtr::dynamicCast(serializer);

        Pose::deserialize(obj);
        frame = obj->getString("frame");
    }


    FramedPosition::FramedPosition() :
        Vector3()
    {
        frame = "";
    }

    FramedPosition::FramedPosition(const Eigen::Vector3f &v, const std::string &s) :
        Vector3(v)
    {
        frame = s;
    }

    FramedPosition::FramedPosition(const Eigen::Matrix4f &m, const std::string &s) :
        Vector3(m)
    {
        frame = s;
    }

    // this doesnt work for unknown reasons
//    FramedPosition::FramedPosition(const Vector3BasePtr pos, const std::string &frame )
//    {
//        this->x = pos->x;
//        this->y = pos->y;
//        this->z = pos->z;
//        this->frame = frame;
//    }

    std::string FramedPosition::getFrame() const
    {
        return this->frame;
    }

    string FramedPosition::output(const Ice::Current &c) const
    {
        std::stringstream s;
        s << toEigen() << std::endl << "frame: " << getFrame();
        return s.str();
    }

    int FramedPosition::readFromXML(const string &xmlData, const Ice::Current &c){
        using namespace boost::property_tree;
        ptree pt;
        std::stringstream stream;
        stream << xmlData;
        xml_parser::read_xml(stream, pt);

        Vector3::readFromXML(xmlData);
        frame = (pt.get<std::string>("frame"));

        return 1;
    }

    string FramedPosition::writeAsXML(const Ice::Current &)
    {
        using namespace boost::property_tree;
        std::stringstream stream;
        stream << Vector3::writeAsXML() <<
                  "<frame>"<<frame<<"</frame>";
        return stream.str();
    }

    void FramedPosition::serialize(const ObjectSerializerBasePtr& serializer, const ::Ice::Current& c) const
    {
        AbstractObjectSerializerPtr obj = AbstractObjectSerializerPtr::dynamicCast(serializer);

        Vector3::serialize(obj, c);
        obj->setString("frame", frame);
    }

    void FramedPosition::deserialize(const ObjectSerializerBasePtr& serializer, const ::Ice::Current& c)
    {
        AbstractObjectSerializerPtr obj = AbstractObjectSerializerPtr::dynamicCast(serializer);

        Vector3::deserialize(obj);
        frame = obj->getString("frame");
    }


    FramedOrientation::FramedOrientation() :
        Quaternion()
    {
        frame = "";
    }

    FramedOrientation::FramedOrientation(const Eigen::Matrix3f &m, const std::string &s) :
        Quaternion(m)
    {
        frame = s;
    }

    FramedOrientation::FramedOrientation(const Eigen::Matrix4f &m, const std::string &s) :
        Quaternion(m)
    {
        frame = s;
    }

    // this doesnt work for an unknown reason
//    FramedOrientation::FramedOrientation(const QuaternionBasePtr ori, const std::string &frame )
//    {
//        Matrix3f rot;
//        rot = Quaternionf(ori->qw, ori->qx, ori->qy, ori->qz);
//        FramedOrientation(rot, frame);
//    }

    std::string FramedOrientation::getFrame() const
    {
        return this->frame;
    }

    string FramedOrientation::output(const Ice::Current &c) const
    {
        std::stringstream s;
        s << toEigen() << std::endl << "frame: " << getFrame();
        return s.str();
    }

    int FramedOrientation::readFromXML(const string &xmlData, const Ice::Current &c)
    {
        using namespace boost::property_tree;
        ptree pt;
        std::stringstream stream;
        stream << xmlData;
        xml_parser::read_xml(stream, pt);

        Quaternion::readFromXML(xmlData);
        frame = (pt.get<std::string>("frame"));

        return 1;
    }

    string FramedOrientation::writeAsXML(const Ice::Current &)
    {
        using namespace boost::property_tree;
        std::stringstream stream;
        stream << Quaternion::writeAsXML() <<
                  "<frame>" << frame << "</frame>";
        return stream.str();
    }

    void FramedOrientation::serialize(const ObjectSerializerBasePtr& serializer, const ::Ice::Current& c) const
    {
        AbstractObjectSerializerPtr obj = AbstractObjectSerializerPtr::dynamicCast(serializer);

        Quaternion::serialize(obj, c);
        obj->setString("frame", frame);
    }

    void FramedOrientation::deserialize(const ObjectSerializerBasePtr& serializer, const ::Ice::Current& c)
    {
        AbstractObjectSerializerPtr obj = AbstractObjectSerializerPtr::dynamicCast(serializer);

        Quaternion::deserialize(obj);
        frame = obj->getString("frame");
    }

    namespace ArmarPose
    {
        VirtualRobot::LinkedCoordinate createLinkedCoordinate(VirtualRobot::RobotPtr &virtualRobot, const FramedPositionPtr& position, const FramedOrientationPtr& orientation)
        {
            VirtualRobot::LinkedCoordinate c(virtualRobot);
            std::string frame = position->getFrame();

            if(!(frame == orientation->getFrame()))
                throw armarx::UserException("Error: frames mismatch");

            Eigen::Matrix4f pose = Eigen::Matrix4f::Identity();

            pose.block<3,3>(0,0) = orientation->toEigen();
            pose.block<3,1>(0,3) = position->toEigen();

            c.set(frame, pose);

            return c;
        }
    }




}
