#include "LinkedPose.h"

#include "RemoteRobot.h"
#include "../SharedRobotServants.h"

#include <Core/core/logging/Logging.h>

#include <boost/property_tree/ptree.hpp>
#include <boost/property_tree/xml_parser.hpp>

#include <Eigen/Geometry>
#include <Eigen/Core>

#include <VirtualRobot/LinkedCoordinate.h>
#include <VirtualRobot/VirtualRobot.h>


namespace armarx {

    LinkedPose::LinkedPose() :
        Pose(),
        FramedPose()
    {
        this->referenceRobot = NULL;
    }

    LinkedPose::LinkedPose(const Eigen::Matrix3f& m, const Eigen::Vector3f& v, const std::string& s, const SharedRobotInterfacePrx& referenceRobot) :
        Pose(m, v),
        FramedPose(m, v, s)
    {
        referenceRobot->ref();
        this->referenceRobot = referenceRobot;
    }

    LinkedPose::LinkedPose(const Eigen::Matrix4f& m, const std::string& s, const SharedRobotInterfacePrx& referenceRobot) :
        Pose(m),
        FramedPose(m, s)
    {
        referenceRobot->ref();
        this->referenceRobot = referenceRobot;
    }

    LinkedPose::~LinkedPose()
    {
        try{
            if(referenceRobot)
            {
                referenceRobot->unref();
            }
        }
        catch(...)
        {
            handleExceptions();
        }
    }


    VirtualRobot::LinkedCoordinate LinkedPose::createLinkedCoordinate()
    {
        VirtualRobot::RobotPtr sharedRobot(new RemoteRobot(referenceRobot));
        VirtualRobot::LinkedCoordinate c(sharedRobot);
        std::string frame = this->getFrame();

        Eigen::Matrix4f pose = Eigen::Matrix4f::Identity();

        pose.block<3,3>(0,0) = QuaternionPtr::dynamicCast(orientation)->toEigen();
        pose.block<3,1>(0,3) = Vector3Ptr::dynamicCast(position)->toEigen();

        c.set(frame, pose);

        return c;
    }


    int LinkedPose::readFromXML(const std::string &xmlData, const Ice::Current &c){
        using namespace boost::property_tree;
        ptree pt;
        std::stringstream stream;
        stream << xmlData;
        xml_parser::read_xml(stream, pt);

        Pose::readFromXML(xmlData);
        frame = (pt.get<std::string>("frame"));

        std::string remoteRobotId = pt.get<std::string>("referenceRobot");
        referenceRobot = SharedRobotInterfacePrx::uncheckedCast(c.adapter->getCommunicator()->stringToProxy(remoteRobotId));
        if(!referenceRobot)
        {
            ARMARX_ERROR_S << "ReferenceRobot for LinkedPose not registered: " << remoteRobotId << flush;
            return 0;
        }

        return 1;
    }

    std::string LinkedPose::writeAsXML(const Ice::Current& c)
    {
        using namespace boost::property_tree;
        std::stringstream stream;
        stream << FramedPose::writeAsXML()
               << "<referenceRobot>" << "Not serializable" <<"</referenceRobot>";
        return stream.str();
    }

    void LinkedPose::serialize(const ObjectSerializerBasePtr& serializer, const ::Ice::Current& c) const
    {
        AbstractObjectSerializerPtr obj = AbstractObjectSerializerPtr::dynamicCast(serializer);

        Pose::serialize(obj, c);
        obj->setString("referenceRobot", "");
    }

    void LinkedPose::deserialize(const ObjectSerializerBasePtr& serializer, const ::Ice::Current& c)
    {
        AbstractObjectSerializerPtr obj = AbstractObjectSerializerPtr::dynamicCast(serializer);

        FramedPose::deserialize(obj);

        std::string remoteRobotId = obj->getString("referenceRobot");
        referenceRobot = SharedRobotInterfacePrx::uncheckedCast(c.adapter->getCommunicator()->stringToProxy(remoteRobotId));
        if(!referenceRobot)
        {
            ARMARX_ERROR_S << "ReferenceRobot for LinkedPose not registered: " << remoteRobotId << flush;
        }
    }

    LinkedVector3::LinkedVector3()
    {
    }

    LinkedVector3::LinkedVector3(const LinkedVector3 &source) :
        IceUtil::Shared(source),
        Vector3Base(source),
        FramedVector3Base(source),
        LinkedVector3Base(source),
        Vector3(source),
        FramedVector3(source)
    {
        referenceRobot = source.referenceRobot;
    }

    LinkedVector3::LinkedVector3(const Eigen::Vector3f &v, const std::string &frame, const SharedRobotInterfacePrx &referenceRobot) :
        FramedVector3(v, frame)
    {
        referenceRobot->ref();
        this->referenceRobot = referenceRobot;
    }

    LinkedVector3::~LinkedVector3()
    {
        try{
            if(referenceRobot)
            {
                referenceRobot->unref();
            }
        }
        catch(...)
        {
            handleExceptions();
        }
    }

    void LinkedVector3::changeFrame(const std::string &newFrame, const Ice::Current &c)
    {
        ARMARX_WARNING_S << "This function doesn't work!";
        
        if(newFrame == frame)
            return;

        VirtualRobot::RobotPtr sharedRobot(new RemoteRobot(referenceRobot));

        FramedVector3Ptr frVec = ChangeFrame(sharedRobot, *this, newFrame);
        x = frVec->x;
        y = frVec->y;
        z = frVec->z;
        frame = frVec->frame;
    }

    int LinkedVector3::readFromXML(const std::string &xmlData, const Ice::Current &c)
    {
        throw LocalException("LinkedVector3 cannot be serialized! Serialize FramedVector3");

    }

    std::string LinkedVector3::writeAsXML(const Ice::Current &c)
    {
        throw LocalException("LinkedVector3 cannot be deserialized! Serialize FramedVector3");

    }

    void LinkedVector3::serialize(const ObjectSerializerBasePtr &serializer, const Ice::Current &) const
    {
        throw LocalException("LinkedVector3 cannot be serialized! Serialize FramedVector3");
    }

    void LinkedVector3::deserialize(const ObjectSerializerBasePtr &serializer, const Ice::Current &)
    {
        throw LocalException("LinkedVector3 cannot be deserialized! Deserialize FramedVector3");
    }





}
