/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotStateComponent::
 * @author     ( stefan dot ulbrich at kit dot edu)
 * @date
 * @copyright  http://www.gnu.org/licenses/gpl.txt
 *             GNU General Public License
 */

#ifndef _ARMARX_RobotState_LinkedPose_H
#define _ARMARX_RobotState_LinkedPose_H

#include "ArmarPose.h"

#include <RobotAPI/interface/robotstate/LinkedPoseBase.h>
#include <RobotAPI/interface/robotstate/RobotState.h>

#include <Core/observers/AbstractObjectSerializer.h>
#include <Core/observers/variant/Variant.h>

#include <Eigen/Core>
#include <Eigen/Geometry>



#include <sstream>


namespace armarx
{
    namespace VariantType
    {
        // variant types
        const VariantTypeId LinkedPose = Variant::addTypeName("::armarx::LinkedPoseBase");
        const VariantTypeId LinkedVector3 = Variant::addTypeName("::armarx::LinkedVector3Base");
    }


    /**
     * @class LinkedPose
     * @ingroup Variants
     * @brief The LinkedPose class
     */
    class LinkedPose :
            virtual public LinkedPoseBase,
            virtual public FramedPose
    {
    public:
        LinkedPose();
        LinkedPose(const Eigen::Matrix3f& m, const Eigen::Vector3f& v, const std::string& frame, const SharedRobotInterfacePrx& referenceRobot);
        LinkedPose(const Eigen::Matrix4f& m, const std::string& frame, const SharedRobotInterfacePrx& referenceRobot);

        virtual ~LinkedPose();

        VirtualRobot::LinkedCoordinate createLinkedCoordinate();


        // inherited from VariantDataClass
        Ice::ObjectPtr ice_clone() const
        {
            return this->clone();
        }

        VariantDataClassPtr clone(const Ice::Current& c = ::Ice::Current()) const
        {
            return new LinkedPose(*this);
        }

        std::string output(const Ice::Current& c = ::Ice::Current()) const
        {
            std::stringstream s;
            s << FramedPose::toEigen() << std::endl << "reference robot: " << referenceRobot;
            return s.str();
        }

        VariantTypeId getType(const Ice::Current& c = ::Ice::Current()) const
        {
            return VariantType::LinkedPose;
        }

        bool validate(const Ice::Current& c = ::Ice::Current())
        {
            return true;
        }

        /**
             * @brief Implementation of virtual function to read a LinkedPose from an XML-file.
             *Example xml-layout:
             * @code
             *      <referenceRobot>5029a22c-e333-4f87-86b1-cd5e0fcce509</referenceRobot>
             *      <frame>leftHand</frame>
             *      <x>0</x>
             *      <y>0</y>
             *      <z>0</z>
             * @endcode

             * @param xmlData String with xml-data. NOT a file path!
             * @return ErrorCode, 1 on Success
             */
        int readFromXML(const std::string &xmlData, const Ice::Current &c = ::Ice::Current());
        std::string writeAsXML(const Ice::Current &c = ::Ice::Current());

        friend std::ostream &operator<<(std::ostream &stream, const LinkedPose& rhs)
        {
            stream << "LinkedPose: " << std::endl << rhs.output() << std::endl;
            return stream;
        };

    public:
        virtual void serialize(const armarx::ObjectSerializerBasePtr& serializer, const ::Ice::Current& = ::Ice::Current()) const;
        virtual void deserialize(const armarx::ObjectSerializerBasePtr& serializer, const ::Ice::Current& = ::Ice::Current());
    };

    typedef IceInternal::Handle<LinkedPose> LinkedPosePtr;

    /**
     * @class LinkedVector3
     * @ingroup Variants
     * @brief The LinkedVector3 class
     */
    class LinkedVector3 :
            virtual public LinkedVector3Base,
            virtual public FramedVector3
    {
    public:
        LinkedVector3();
        LinkedVector3(const LinkedVector3& source);
        LinkedVector3(const Eigen::Vector3f& v, const std::string& frame, const SharedRobotInterfacePrx& referenceRobot);

        virtual ~LinkedVector3();

        void changeFrame(const std::string &newFrame, const Ice::Current &c = Ice::Current());

        // inherited from VariantDataClass
        Ice::ObjectPtr ice_clone() const
        {
            return this->clone();
        }

        VariantDataClassPtr clone(const Ice::Current& c = ::Ice::Current()) const
        {
            return new LinkedVector3(*this);
        }

        std::string output(const Ice::Current& c = ::Ice::Current()) const
        {
            std::stringstream s;
            s << FramedVector3::toEigen() << std::endl << "reference robot: " << referenceRobot;
            return s.str();
        }

        VariantTypeId getType(const Ice::Current& c = ::Ice::Current()) const
        {
            return VariantType::LinkedVector3;
        }

        bool validate(const Ice::Current& c = ::Ice::Current())
        {
            return true;
        }


        int readFromXML(const std::string &xmlData, const Ice::Current &c = ::Ice::Current());
        std::string writeAsXML(const Ice::Current &c = ::Ice::Current());

        friend std::ostream &operator<<(std::ostream &stream, const LinkedVector3& rhs)
        {
            stream << "LinkedVector3: " << std::endl << rhs.output() << std::endl;
            return stream;
        };

        virtual void serialize(const armarx::ObjectSerializerBasePtr& serializer, const ::Ice::Current& = ::Ice::Current()) const;
        virtual void deserialize(const armarx::ObjectSerializerBasePtr& serializer, const ::Ice::Current& = ::Ice::Current());

    };
    typedef IceInternal::Handle<LinkedVector3> LinkedVector3Ptr;
}

#endif
