/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as
* published by the Free Software Foundation; either version 2 of
* the License, or (at your option) any later version.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU Lesser General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @package    ArmarX::Core
* @author     Stefan Ulbrich <stefan dot ulbrich at kit dot edu>, Kai Welke (welke _at_ kit _dot_ edu)
* @date       2011 Kai Welke
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

#include "RobotStateObserver.h"
#include <RobotAPI/libraries/robotstate/RobotStateComponent.h>
#include <Core/observers/checks/ConditionCheckEquals.h>
#include <Core/observers/checks/ConditionCheckInRange.h>
#include <Core/observers/checks/ConditionCheckLarger.h>
#include <Core/observers/checks/ConditionCheckSmaller.h>

#include <RobotAPI/libraries/robotstate/remote/RemoteRobot.h>

#include <VirtualRobot/VirtualRobot.h>
#include <VirtualRobot/RobotNodeSet.h>
#include <VirtualRobot/Nodes/RobotNode.h>
#include <VirtualRobot/VirtualRobot.h>


using namespace armarx;

// ********************************************************************
// observer framework hooks
// ********************************************************************
void RobotStateObserver::onInitObserver()
{
	// read names of kinematic chain elements belonging to this unit from XML and setup a map of all joints
    // the kinematic chain elements belonging to this unit are defined in a RobotNodeSet
    std::string robotFile = getProperty<std::string>("RobotFileName").getValue();
    robotNodeSetName = getProperty<std::string>("RobotNodeSetName").getValue();

	this->server = getProxy<RobotStateComponentInterfacePrx>("RobotStateComponent");
	this->robot.reset(new RemoteRobot(server->getSynchronizedRobot()));

    if(robotNodeSetName == "")
    {
        throw UserException("RobotNodeSet not defined");
    }
    VirtualRobot::RobotNodeSetPtr robotNodeSetPtr = this->robot->getRobotNodeSet(robotNodeSetName);

    std::vector< VirtualRobot::RobotNodePtr > robotNodes;
    robotNodes = robotNodeSetPtr->getAllRobotNodes();


	// register all channels
	offerChannel("positions","Link positions of the " + robotNodeSetName + " kinematic chain");
	offerChannel("orientations","Link orientations of the" + robotNodeSetName + "kinematic chain");


	// register all data fields
    for (std::vector<VirtualRobot::RobotNodePtr>::iterator it = robotNodes.begin(); it != robotNodes.end(); it++)
    {
        std::string jointName = (*it)->getName();
        std::cout << "* " << jointName << std::endl;
  		offerDataField("positions", jointName, VariantType::FramedPosition, "" + jointName + "");
  		offerDataField("orientations", jointName, VariantType::FramedOrientation, "" + jointName + "");

  		// create lookup map
  		namesToNodes.insert(make_pair(jointName, *it));
    }

	// register all checks
	offerConditionCheck("equals", new ConditionCheckEquals());
	offerConditionCheck("inrange", new ConditionCheckInRange());
	offerConditionCheck("larger", new ConditionCheckLarger());
	offerConditionCheck("smaller", new ConditionCheckSmaller());

	usingTopic(robotNodeSetName + "State");

}

void RobotStateObserver::onConnectObserver()
{
}

// ********************************************************************
// KinematicUnitListener interface implementation
// ********************************************************************
void RobotStateObserver::reportControlModeChanged(const NameControlModeMap& jointModes, bool aValueChanged, const Ice::Current& c)
{
}

void RobotStateObserver::reportJointAngles(const NameValueMap& jointAngles, bool aValueChanged, const Ice::Current& c)
{
    if(!aValueChanged)
        return;
	// update current model
	updateRobotModel(jointAngles);

	// update datafields
	updatePoses();

	// update channels
	updateChannel("positions");
	updateChannel("orientations");
}

void RobotStateObserver::reportJointVelocities(const NameValueMap& jointVelocities, bool aValueChanged, const Ice::Current& c)
{
}

void RobotStateObserver::reportJointTorques(const NameValueMap& jointTorques, bool aValueChanged, const Ice::Current& c)
{
}

void RobotStateObserver::reportJointCurrents(const NameValueMap& jointCurrents, bool aValueChanged, const Ice::Current& c)
{
}

void RobotStateObserver::reportJointMotorTemperatures(const NameValueMap& jointMotorTemperatures, bool aValueChanged, const Ice::Current& c)
{
}

void RobotStateObserver::reportJointStatuses(const NameStatusMap &jointStatuses, bool aValueChanged, const Ice::Current &c)
{
}

// ********************************************************************
// private methods
// ********************************************************************
void RobotStateObserver::updateRobotModel(const NameValueMap& nameValueMap)
{
	// update all nodes contained in the valueMap
	NameValueMap::const_iterator iter = nameValueMap.begin();

	while(iter != nameValueMap.end())
	{
	    VirtualRobot::RobotNodePtr robotNode = namesToNodes[iter->first];
	    robotNode->setJointValue(iter->second);

		iter++;
	}
}

void RobotStateObserver::updatePoses()
{
	// go through all nodes
	VirtualRobot::RobotNodeSetPtr robotNodeSetPtr = this->robot->getRobotNodeSet(robotNodeSetName);

    std::vector< VirtualRobot::RobotNodePtr > robotNodes;
    robotNodes = robotNodeSetPtr->getAllRobotNodes();

	// set all data fields
	// TODO: maybe global pose is not correct!!!
    for (std::vector<VirtualRobot::RobotNodePtr>::iterator it = robotNodes.begin(); it != robotNodes.end(); it++)
    {
        std::string jointName = (*it)->getName();
        setDataField("positions", jointName, FramedPositionPtr(new FramedPosition((*it)->getGlobalPose(), "global")));
        setDataField("orientations", jointName, FramedOrientationPtr(new FramedOrientation((*it)->getGlobalPose(), "global")));
    }
}

PropertyDefinitionsPtr RobotStateObserver::createPropertyDefinitions()
{
    return PropertyDefinitionsPtr(new RobotStatePropertyDefinitions(
                                           getConfigIdentifier()));
}
