/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as
* published by the Free Software Foundation; either version 2 of
* the License, or (at your option) any later version.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU Lesser General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @package    ArmarX::Core
* @author     Stefan Ulbrich <stefan dot ulbrich at kit dot edu>, Kai Welke (welke _at_ kit _dot_ edu)
* @date       2011 Kai Welke
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

#ifndef _ARMARX_CORE_KINEMATIC_UNIT_OBSERVER_H
#define _ARMARX_CORE_KINEMATIC_UNIT_OBSERVER_H

#include <Core/core/system/ImportExport.h>
#include <Core/core/Component.h>
#include <Core/interface/observers/VariantBase.h>
#include <RobotAPI/interface/units/KinematicUnitInterface.h>
#include <RobotAPI/interface/robotstate/RobotStateObserverInterface.h>
#include <RobotAPI/interface/robotstate/RobotState.h>
#include <Core/observers/ConditionCheck.h>
#include <Core/observers/Observer.h>

#include <string>

namespace VirtualRobot
{
    class Robot;
    typedef boost::shared_ptr<Robot> RobotPtr;
    class RobotNode;
    typedef boost::shared_ptr<RobotNode> RobotNodePtr;
}

namespace armarx
{

   /**
    * ArmarX RobotStateObserver.
    *
    * The RobotStateObserver allows to install conditions on all channel reported by the KinematicUnit.
    * These include joint angles, velocities, torques and motor temperatures
    *
    * The RobotStateObserver retrieves its configuration from a VirtualRobot robot model. Within the model, the joints
    * which are observer by the unit are define by a robotnodeset
    */
    class ARMARXCORE_IMPORT_EXPORT RobotStateObserver :
        virtual public Observer,
        virtual public RobotStateObserverInterface
    {
    public:
   		// framework hooks
   		void onInitObserver();
   		void onConnectObserver();

		// slice interface implementation
        void reportControlModeChanged(const NameControlModeMap& jointModes, bool aValueChanged, const Ice::Current& c = ::Ice::Current());
        void reportJointAngles(const NameValueMap& jointAngles, bool aValueChanged, const Ice::Current& c = ::Ice::Current());
        void reportJointVelocities(const NameValueMap& jointVelocities, bool aValueChanged, const Ice::Current& c = ::Ice::Current());
        void reportJointTorques(const NameValueMap& jointTorques, bool aValueChanged, const Ice::Current& c = ::Ice::Current());
        void reportJointCurrents(const NameValueMap& jointCurrents, bool aValueChanged, const Ice::Current& c = ::Ice::Current());
        void reportJointMotorTemperatures(const NameValueMap& jointMotorTemperatures, bool aValueChanged, const Ice::Current& c = ::Ice::Current());
        void reportJointStatuses(const NameStatusMap& jointStatuses, bool aValueChanged, const Ice::Current& c = Ice::Current());

        virtual PropertyDefinitionsPtr createPropertyDefinitions();

        virtual std::string getDefaultName() const { return "RobotStateObserver"; }

    protected:
		void updateRobotModel(const NameValueMap& nameValueMap);
		void updatePoses();

    private:
    	std::string robotNodeSetName;
		RobotStateComponentInterfacePrx server;
		VirtualRobot::RobotPtr  robot;
		std::map<std::string,VirtualRobot::RobotNodePtr> namesToNodes;
    };
}

#endif
