/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    ArmarXCore::units
 * @author     Manfred Kroehnert (Manfred dot Kroehnert at kit dot edu)
 * @date       2013
 * @copyright  http://www.gnu.org/licenses/gpl.txt
 *             GNU General Public License
 */


#include "HandUnit.h"

#include <Core/core/system/ArmarXDataPath.h>

#include <VirtualRobot/VirtualRobot.h>
#include <VirtualRobot/EndEffector/EndEffector.h>
#include <VirtualRobot/XML/RobotIO.h>

#include <vector>

using namespace armarx;
using namespace VirtualRobot;


void HandUnit::onInitComponent()
{
    std::string endeffectorFile = getProperty<std::string>("RobotFileName").getValue();
    std::string endeffectorName = getProperty<std::string>("EndeffectorName").getValue();

    if (!ArmarXDataPath::getAbsolutePath(endeffectorFile, endeffectorFile))
    {
        throw UserException("Could not find robot file " + endeffectorFile);
    }


    try
    {
        robot = VirtualRobot::RobotIO::loadRobot(endeffectorFile, VirtualRobot::RobotIO::eStructure);
    }
    catch(VirtualRobot::VirtualRobotException& e)
    {
        throw UserException(e.what());
    }

    if(endeffectorName == "")
    {
        throw UserException("EndeffectorName not defined");
    }
    if(!robot->hasEndEffector(endeffectorName))
    {
        throw UserException("Robot does not contain an endeffector with name: " + endeffectorName);
    }
    eef = robot->getEndEffector(endeffectorName);
    robotName = robot->getName();
    if (eef->getTcp())
        tcpName = robot->getEndEffector(endeffectorName)->getTcp()->getName();
    else
        throw UserException("Endeffector without TCP: " + endeffectorName);

    // get all joints
    std::vector<EndEffectorActorPtr> actors;
    eef->getActors(actors);
    for (size_t i=0;i<actors.size();i++)
    {
        EndEffectorActorPtr a = actors[i];
        ARMARX_WARNING_S << "here is a hack to make the stuff compile - but it will probably crash when run!";
        std::vector<EndEffectorActor::ActorDefinition> ads;// = a->getDefinition();
        for (size_t j=0;j<ads.size();j++)
        {
            EndEffectorActor::ActorDefinition ad = ads[j];
            if (ad.directionAndSpeed!=0 && ad.robotNode)
            {
                handJoints[ad.robotNode->getName()] = ad.directionAndSpeed;
            }
        }
    }

    const std::vector<std::string> endeffectorPreshapes = robot->getEndEffector(endeffectorName)->getPreshapes();

    preshapeNames = new SingleTypeVariantList(VariantType::String);
    std::vector<std::string>::const_iterator iter = endeffectorPreshapes.begin();
    while(iter != endeffectorPreshapes.end())
    {
        Variant currentPreshape;
        currentPreshape.setString(*iter);
        preshapeNames->addVariant(currentPreshape);
        iter++;
    }


    // component dependencies
    listenerChannelName = endeffectorName + "State";
    offeringTopic(listenerChannelName);

    this->onInitHandUnit();
}


void HandUnit::onConnectComponent()
{
    ARMARX_INFO << "setting report topic to " << listenerChannelName << flush;
    listenerPrx = getTopic<HandUnitListenerPrx>(listenerChannelName);

    this->onStartHandUnit();
}


void HandUnit::onExitComponent()
{
    this->onExitHandUnit();
}


void HandUnit::open(const Ice::Current& c)
{
}


void HandUnit::close(const Ice::Current& c)
{
}


void HandUnit::preshape(const std::string& preshapeName, const Ice::Current& c)
{
}


SingleTypeVariantListBasePtr HandUnit::getPreshapeNames(const Ice::Current& c)
{
    return preshapeNames;
}


PropertyDefinitionsPtr HandUnit::createPropertyDefinitions()
{
    return PropertyDefinitionsPtr(new HandUnitPropertyDefinitions(
                                           getConfigIdentifier()));
}


NameValueMap HandUnit::getPreshapeJointValues(const std::string &preshapeName, const Ice::Current &)
{
    EndEffectorPtr efp = robot->getEndEffector(getProperty<std::string>("EndeffectorName").getValue());
    RobotConfigPtr rc = efp->getPreshape(preshapeName);
    return rc->getRobotNodeJointValueMap();
}

void HandUnit::setObjectGrasped(const std::string &objectName, const Ice::Current &)
{
   ARMARX_INFO << "Object grasped " << objectName << flush;
   graspedObject = objectName;
}

void HandUnit::setObjectReleased(const std::string &objectName, const Ice::Current &)
{
    ARMARX_INFO << "Object released " << objectName << flush;
    graspedObject = "";
}

