#include "HeadIKUnit.h"


#include <Core/core/system/ArmarXDataPath.h>

#include <VirtualRobot/XML/RobotIO.h>
#include <VirtualRobot/IK/GazeIK.h>
#include <VirtualRobot/LinkedCoordinate.h>


namespace armarx
{

    HeadIKUnit::HeadIKUnit() :
        requested(false),
        cycleTime(30),
        newTargetSet(false)
    {
        targetPosition = new FramedPosition();
    }


    void HeadIKUnit::onInitComponent()
    {
        ScopedLock lock(accessMutex);

        usingProxy(getProperty<std::string>("KinematicUnitName").getValue());
        usingProxy("RobotStateComponent");
        usingTopic("RobotState");

        cycleTime = getProperty<int>("CycleTime").getValue();
    }


    void HeadIKUnit::onConnectComponent()
    {
        ScopedLock lock(accessMutex);

        kinematicUnitPrx = getProxy<KinematicUnitInterfacePrx>(getProperty<std::string>("KinematicUnitName").getValue());
        robotStateComponentPrx = getProxy<RobotStateComponentInterfacePrx>("RobotStateComponent");

        //remoteRobotPrx = robotStateComponentPrx->getSynchronizedRobot();
        localRobot = RemoteRobot::createLocalClone(robotStateComponentPrx);


        //std::string robotModelFile;
        //ArmarXDataPath::getAbsolutePath("Armar3/data/robotmodel/ArmarIII.xml", robotModelFile);
        //localRobot = VirtualRobot::RobotIO::loadRobot(robotModelFile.c_str(), VirtualRobot::RobotIO::eStructure);
        //VirtualRobot::RobotPtr robotSnapshot(new RemoteRobot(remoteRobotPrx));
        //localRobot->setConfig(robotSnapshot->getConfig());

        requested = false;
        if (execTask) execTask->stop();
        execTask = new PeriodicTask<HeadIKUnit>(this, &HeadIKUnit::periodicExec, cycleTime, false, "HeadIKCalculator");
        execTask->setDelayWarningTolerance(300);
        execTask->start();
    }

    void HeadIKUnit::onDisconnectComponent()
    {
        ScopedLock lock(accessMutex);

        release();
        if (execTask) execTask->stop();
    }

    void HeadIKUnit::onExitComponent()
    {
    }



    void HeadIKUnit::setCycleTime(Ice::Int milliseconds, const Ice::Current & c)
    {
        ScopedLock lock(accessMutex);

        cycleTime = milliseconds;
        if (execTask) execTask->changeInterval(cycleTime);
    }


    void HeadIKUnit::setHeadTarget(const std::string& robotNodeSetName, const FramedVector3BasePtr& targetPosition, const Ice::Current& c)
    {
        ScopedLock lock(accessMutex);

        this->robotNodeSetName = robotNodeSetName;
        this->targetPosition->x = targetPosition->x;
        this->targetPosition->y = targetPosition->y;
        this->targetPosition->z = targetPosition->z;
        this->targetPosition->frame = targetPosition->frame;
        newTargetSet = true;
    }




    void HeadIKUnit::init(const Ice::Current & c)
    {
    }

    void HeadIKUnit::start(const Ice::Current & c)
    {
    }

    void HeadIKUnit::stop(const Ice::Current & c)
    {
    }

    UnitExecutionState HeadIKUnit::getExecutionState(const Ice::Current & c)
    {
        switch (getState())
        {
            case eManagedIceObjectStarted:
                return eUnitStarted;
            case eManagedIceObjectInitialized:
            case eManagedIceObjectStarting:
                return eUnitInitialized;
            case eManagedIceObjectExiting:
            case eManagedIceObjectExited:
                return eUnitStopped;
            default:
                return eUnitConstructed;
        }
    }




    void HeadIKUnit::request(const Ice::Current & c)
    {
        ScopedLock lock(accessMutex);

        requested = true;
        ARMARX_IMPORTANT << "Requesting HeadIKUnit";
        if(execTask) execTask->stop();
        execTask = new PeriodicTask<HeadIKUnit>(this, &HeadIKUnit::periodicExec, cycleTime, false, "TCPVelocityCalculator");
        execTask->start();
        ARMARX_IMPORTANT << "Requested HeadIKUnit";
    }




    void HeadIKUnit::release(const Ice::Current & c)
    {
        ScopedLock lock(accessMutex);

        ARMARX_IMPORTANT << "Releasing HeadIKUnit";
        requested = false;
        kinematicUnitPrx->stop();
        ARMARX_IMPORTANT << "Released HeadIKUnit";
    }




    void HeadIKUnit::periodicExec()
    {
        bool doCalculation = false;

        {
            ScopedLock lock(accessMutex);

            doCalculation = requested && newTargetSet;
            newTargetSet = false;
        }


        if (doCalculation)
        {
            RemoteRobot::synchronizeLocalClone(localRobot,robotStateComponentPrx);
            //VirtualRobot::RobotPtr robotSnapshot(new RemoteRobot(remoteRobotPrx));
            //localRobot->setConfig(robotSnapshot->getConfig());

            VirtualRobot::RobotNodeSetPtr kinematicChain = localRobot->getRobotNodeSet(robotNodeSetName);
            VirtualRobot::RobotNodePrismaticPtr virtualJoint;
            for (unsigned int i=0; i<kinematicChain->getSize(); i++)
            {
                if (kinematicChain->getNode(i)->getName().compare("VirtualCentralGaze") == 0)
                {
                    virtualJoint = boost::dynamic_pointer_cast<VirtualRobot::RobotNodePrismatic>(kinematicChain->getNode(i));
                }
            }
            VirtualRobot::GazeIK ikSolver(kinematicChain, virtualJoint);
            ikSolver.enableJointLimitAvoidance(true);

            Eigen::Matrix3f m = Eigen::Matrix3f::Identity();
            FramedOrientationPtr targetOri = new FramedOrientation(m, targetPosition->frame);
            VirtualRobot::LinkedCoordinate target = ArmarPose::createLinkedCoordinate(localRobot, targetPosition, targetOri);
            Eigen::Matrix4f goalInRoot = target.getInFrame(localRobot->getRootNode());
            Eigen::Vector3f targetPos = goalInRoot.block<3,1>(0,3);

            ARMARX_VERBOSE << "Calculating IK for target position " << targetPos;
            if (!ikSolver.solve(targetPos))
            {
                ARMARX_WARNING << "IKSolver found no solution!";
            }
            else
            {
                NameValueMap targetJointAngles;
                for (int i = 0; i < (signed int)kinematicChain->getSize(); i++)
                {
                    if (kinematicChain->getNode(i)->getName().compare("VirtualCentralGaze") != 0)
                    {
                        targetJointAngles[kinematicChain->getNode(i)->getName()] = kinematicChain->getNode(i)->getJointValue();
                    }
                }
                kinematicUnitPrx->setJointAngles(targetJointAngles);
            }
        }
    }




    PropertyDefinitionsPtr HeadIKUnit::createPropertyDefinitions()
    {
        return PropertyDefinitionsPtr(new HeadIKUnitPropertyDefinitions(getConfigIdentifier()));
    }



}
