/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    ArmarXCore::units
 * @author     Christian Boege (boege dot at kit dot edu)
 * @date       2011
 * @copyright  http://www.gnu.org/licenses/gpl.txt
 *             GNU General Public License
 */

#ifndef _ARMARX_COMPONENT_TCPMOVER_UNIT_H
#define _ARMARX_COMPONENT_TCPMOVER_UNIT_H

#include "KinematicUnit.h"

#include <Core/core/Component.h>
#include <Core/core/services/tasks/PeriodicTask.h>
#include <Core/core/application/properties/Properties.h>
#include <Core/core/system/ImportExportComponent.h>
#include <RobotAPI/interface/units/TCPMoverUnitInterface.h>

//
#include <IceUtil/Time.h>

//
#include <Eigen/Core>

#ifndef WIN32
#define KINEMATIC_UNIT_FILE_DEFAULT std::string(std::getenv("ArmarXHome_DIR")?std::getenv("ArmarXHome_DIR"):"../../")+std::string("/data/ArmarIV/RobotModel/ArmarIV.xml")
#else
#define KINEMATIC_UNIT_FILE_DEFAULT "/org/share/staff/home/waechter/projects/armarx/trunk/data/ArmarIV/RobotModel/ArmarIV.xml"
#endif
#define KINEMATIC_UNIT_NAME_DEFAULT "RobotKinematicUnit"

namespace VirtualRobot {
    class Robot;
    class RobotNodeSet;
    class DifferentialIK;
    typedef boost::shared_ptr<Robot> RobotPtr;
    typedef boost::shared_ptr<RobotNodeSet> RobotNodeSetPtr;
    typedef boost::shared_ptr<DifferentialIK> DifferentialIKPtr;

}

namespace armarx
{

    /**
     * @class TCPMoverUnitPropertyDefinitions
     * @brief
     * @ingroup SensorActorUnits
     */
    class TCPMoverUnitPropertyDefinitions:
            public ComponentPropertyDefinitions
    {
    public:
        TCPMoverUnitPropertyDefinitions(std::string prefix):
            ComponentPropertyDefinitions(prefix)
        {
            defineOptionalProperty<std::string>("RobotNodeSetName", KINEMATIC_UNIT_NAME_DEFAULT, "Robot node name");
            defineOptionalProperty<std::string>("RobotFileName", KINEMATIC_UNIT_FILE_DEFAULT, "Robot file name, e.g. robot_model.xml");
        }
    };


    /**
      @class TCPMoverUnit
      @brief
      @ingroup SensorActorUnits
     */
    class TCPMoverUnit :
        virtual public Component,
        virtual public TCPMoverUnitInterface
     {
    public:
        struct HandData{
            HandSelection hand;
            std::vector<float> velocities;
            std::vector<float> orientation;
            Eigen::Matrix4f standardRot;
            VirtualRobot::DifferentialIKPtr ik;


            HandData():
                hand(eNone),
                velocities(3, 0.f),
                orientation(3, 0.f)
            {}

            void setTCPNodeSet( VirtualRobot::RobotNodeSetPtr tcpNodeSet);
            VirtualRobot::RobotNodeSetPtr getTCPNodeSet(){ return tcpNodeSet;}

        protected:
              VirtualRobot::RobotNodeSetPtr tcpNodeSet;

        };

        TCPMoverUnit();
        // inherited from Component
        virtual std::string getDefaultName() const { return "TCPMoverUnit"; }

        virtual void onInitComponent();
        virtual void onConnectComponent();
        virtual void onExitComponent();

        virtual void periodicExec();

        void request(const Ice::Current &);
        void release(const Ice::Current &);
        void resetArmToHomePosition(HandSelection selectedHand, const Ice::Current &);
        void lookWithHeadTo(Ice::Float x, Ice::Float y, Ice::Float z, const Ice::Current &);
        void setCartesianTCPVelocity(HandSelection selectedHand, Ice::Float x, Ice::Float y, Ice::Float z, Ice::Float speedFactor, const Ice::Current &);
        void setTCPOrientation(HandSelection selectedHand, Ice::Float alpha, Ice::Float beta, Ice::Float gamma, const Ice::Current &);
        void setPlatformVelocities(Ice::Float x, Ice::Float y, Ice::Float theta, const Ice::Current &);
        void setRotationMat(Eigen::Matrix4f &matrix, float alpha, float beta, float gamma);
        void resetToHomePosition(HandSelection selectedHand);
        void setHandConfiguration(HandSelection, Ice::Int, const Ice::Current &);


        void stopArm(HandSelection, const Ice::Current &){}
        void stopHead(const Ice::Current &){}
        void stopPlatform(const Ice::Current &){}
        virtual PropertyDefinitionsPtr createPropertyDefinitions();

    protected:
        void calculateNewHandPosition(HandSelection selectedHand);

    private:
        std::map<HandSelection, HandData> handData;
        IceUtil::Time lastExec;
        HandSelection selectedHand;
        NameValueMap velocityMap;
        bool robotRequested;

        std::string kinematicUnitName;
        std::string kinematicUnitFile;
        VirtualRobot::RobotPtr robot;
        KinematicUnitInterfacePrx kinematicUnitPrx;
        PeriodicTask<TCPMoverUnit>::pointer_type periodicTask;
    };
}

#endif
