/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::MotionControlGroup
 * @author     Mirko Waechter ( mirko dot waechter at kit dot edu )
 * @date       2014
 * @copyright  http://www.gnu.org/licenses/gpl.txt
 *             GNU General Public License
 */

#include "MoveJoints.h"

#include <RobotAPI/interface/units/KinematicUnitInterface.h>

#include <RobotAPI/libraries/core/RobotStatechartContext.h>

using namespace armarx;
using namespace MotionControlGroup;

// DO NOT EDIT NEXT LINE
MoveJoints::SubClassRegistry MoveJoints::Registry(MoveJoints::GetName(), &MoveJoints::CreateInstance);



MoveJoints::MoveJoints(XMLStateConstructorParams stateData) :
    XMLStateTemplate < MoveJoints > (stateData)
{
}

void MoveJoints::onEnter()
{
    // put your user code for the enter-point here
    // execution time should be short (<100ms)

    RobotStatechartContext* context = getContext<RobotStatechartContext>();

    NameValueMap targetJointAngles;
    NameValueMap targetJointVelocities;
    SingleTypeVariantListPtr jointNames = getInput<SingleTypeVariantList>("jointNames");
    SingleTypeVariantListPtr targetJointValues = getInput<SingleTypeVariantList>("targetJointValues");
    NameControlModeMap controlModes;
    if (jointNames->getSize()!=targetJointValues->getSize())
    {
        throw LocalException("Sizes of joint name list and joint value list do not match!");
    }

    ARMARX_VERBOSE << "number of joints that will be set: " << jointNames->getSize() << flush;

    for (int i=0; i<jointNames->getSize(); i++)
    {
        targetJointAngles[jointNames->getVariant(i)->getString()] = targetJointValues->getVariant(i)->getFloat();
        controlModes[jointNames->getVariant(i)->getString()] = ePositionControl;
        ARMARX_VERBOSE << "setting joint angle for joint '" << jointNames->getVariant(i)->getString() << "' to " << targetJointValues->getVariant(i)->getFloat() << std::endl;
    }

    // TODO: Set Max Velocities
    if(isInputParameterSet("jointMaxSpeeds") && getInput<SingleTypeVariantList>("jointMaxSpeeds")->getSize() == jointNames->getSize())
    {
        SingleTypeVariantListPtr maxJointValues = getInput<SingleTypeVariantList>("jointMaxSpeeds");
        for (int i=0; i<jointNames->getSize(); i++)
        {
            targetJointVelocities[jointNames->getVariant(i)->getString()] = maxJointValues->getVariant(i)->getFloat();
        }
    }
    else
    {
        float maxVel = getInput<float>("jointMaxSpeed");
        for (int i=0; i<jointNames->getSize(); i++)
        {
            targetJointVelocities[jointNames->getVariant(i)->getString()] = maxVel;
        }
    }

    // now install the condition
    Term cond;
    float tolerance = getInput<float>("jointTargetTolerance");
    for (int i=0; i<jointNames->getSize(); i++)
    {
        ARMARX_VERBOSE << "creating condition (" << i << " of " << jointNames->getSize() << ")" << flush;
        ParameterList inrangeCheck;
        inrangeCheck.push_back(new Variant(targetJointValues->getVariant(i)->getFloat()-tolerance));
        inrangeCheck.push_back(new Variant(targetJointValues->getVariant(i)->getFloat()+tolerance));

        Literal inrangeLiteral(DataFieldIdentifier(context->getKinematicUnitObserverName(), "jointangles", jointNames->getVariant(i)->getString()), "inrange", inrangeCheck);
        cond = cond && inrangeLiteral;
    }

    ARMARX_VERBOSE << "installing condition: EvJointTargetReached" << std::endl;
    targetReachedCondition = installCondition<EvJointTargetReached>(cond);
    ARMARX_VERBOSE << "condition installed: EvJointTargetReached" << std::endl;
    timeoutEvent = setTimeoutEvent(getInput<int>("timeoutInMs"), createEvent<EvTimeout>());
    ARMARX_VERBOSE << "timeout installed" << std::endl;

    context->kinematicUnitPrx->switchControlMode(controlModes);
    context->kinematicUnitPrx->setJointVelocities(targetJointVelocities);
    context->kinematicUnitPrx->setJointAngles(targetJointAngles);
}



void MoveJoints::onBreak()
{
    // put your user code for the breaking point here
    // execution time should be short (<100ms)
}

void MoveJoints::onExit()
{
    // put your user code for the exit point here
    // execution time should be short (<100ms)
    removeTimeoutEvent(timeoutEvent);
    removeCondition(targetReachedCondition);

}

// DO NOT EDIT NEXT FUNCTION
std::string MoveJoints::GetName()
{
    return "MoveJoints";
}

// DO NOT EDIT NEXT FUNCTION
XMLStateFactoryBasePtr MoveJoints::CreateInstance(XMLStateConstructorParams stateData)
{
    return XMLStateFactoryBasePtr(new MoveJoints(stateData));
}

