/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2 of
* the License, or (at your option) any later version.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU Lesser General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @package    RobotAPI::MovePlatformToLandmark
* @author     Valerij Wittenbeck
* @date       2014 Valerij Wittenbeck
* @copyright  http://www.gnu.org/licenses/gpl.txt
*             GNU General Public License
*/

#ifndef ARMARX_COMPONENT_MOVE_PLATFORM_TO_LANDMARK_H
#define ARMARX_COMPONENT_MOVE_PLATFORM_TO_LANDMARK_H

#include <RobotAPI/statecharts/MovePlatform/PlatformContext.h>
#include <Core/statechart/Statechart.h>

/*
* This statechart is meant to bridge the gap between a user verbally specifying a target location (i.e. fridge, table, ...)
* and the robot knowing which points it needs to go to in order to reach that landmark.
*
* Ideally, there is some planning involved (graph retrieving/generating, path searching, etc), but this basic form only associates
* a landmark tag with a series of points that need to be traversed through without respect to the robot's current position.
*/

namespace armarx
{
    class MovePlatformToLandmarkStateChart : public RemoteStateOfferer<PlatformContext>
    {
    public:
        void onInitRemoteStateOfferer();
        void onConnectRemoteStateOfferer();
        std::string getStateOffererName() const { return "MovePlatformToLandmarkStateChart"; }
    };

    // ****************************************************************
    // Events
    // ****************************************************************

    DEFINEEVENT(EvPathFound)
    DEFINEEVENT(EvNoPathFound)

    // ****************************************************************
    // Definition of StatechartMovePlatformToLandmark
    // ****************************************************************
    struct StatechartMovePlatformToLandmark :
        StateTemplate<StatechartMovePlatformToLandmark>
    {
        void defineParameters();
        void defineSubstates();
        void onEnter();
        void onExit();
    };

    // ****************************************************************
    // Definition of StateCalculatePath
    // ****************************************************************
    /**
     * StateCalculatePath: Calculate set of points for the robot to drive through
     */

    struct StateCalculatePath :
        StateTemplate<StateCalculatePath>
    {
        struct Node {
            std::string name;
            std::vector<boost::shared_ptr<Node> > successors;
            FramedVector3Ptr framedPos;
        };
        typedef boost::shared_ptr<Node> NodePtr;

        void defineParameters();
        void onEnter();
        void onExit();

        inline float getDist(float x1, float y1, float x2, float y2) {
            const float dX = x1 - x2;
            const float dY = y1 - y2;
            return std::sqrt(dX * dX + dY * dY);
        }

        std::list<NodePtr> aStar(NodePtr start, NodePtr goal);
        inline float heuristic(NodePtr n1, NodePtr n2) {
            return getDist(n1->framedPos->x, n1->framedPos->y, n2->framedPos->x, n2->framedPos->y);
        }
    };

}

#endif
