/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2 of
* the License, or (at your option) any later version.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU Lesser General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @package    RobotAPI::PlaceObject
* @author     Nikolaus Vahrenkamp
* @date       2014

* @copyright  http://www.gnu.org/licenses/gpl.txt
*             GNU General Public License
*/

#include "PlaceObject.h"
#include <RobotAPI/libraries/core/RobotStatechartContext.h>

#include <RobotAPI/interface/units/KinematicUnitInterface.h>
#include <RobotAPI/libraries/core/RobotStatechartContext.h>
#include <RobotAPI/libraries/robotstate/remote/ArmarPose.h>
#include <RobotAPI/libraries/robotstate/remote/RemoteRobot.h>

#include <Core/observers/variant/ChannelRef.h>
#include <Core/observers/variant/SingleTypeVariantList.h>

#include <VirtualRobot/IK/DifferentialIK.h>
#include <VirtualRobot/Robot.h>


namespace armarx
{
    // ****************************************************************
    // Implementation of StatechartPlaceObject
    // ****************************************************************
    void StatechartPlaceObject::defineParameters()
    {
        addToInput("maxHeightToMoveDown", VariantType::Float, false);
        addToInput("minForceForSuccess", VariantType::Float, false);
        addToInput("timeoutPlaceObject", VariantType::Int, false);

        addToInput("robotNodeSetName", VariantType::String, false);
        addToInput("robotBaseFrame", VariantType::String, false);


        addToLocal("startPose", VariantType::FramedPose);
        addToLocal("goalPose", VariantType::FramedPose);

        //addToLocal("jointNames", VariantType::List(VariantType::String));
        //addToLocal("jointVelocitiesDatafields", VariantType::List(VariantType::DatafieldRef));
    }

    void StatechartPlaceObject::defineSubstates()
    {
        StatePtr stateMoveDown= addState<StateMoveDown>("stateMoveDown");
        //StatePtr stateCheckForces = addState<StateCheckForces>("stateCheckForces");
        StatePtr stateSuccess = addState<SuccessState>("stateSuccess");
        StatePtr stateFailure = addState<FailureState>("stateFailure");

        ParameterMappingPtr mapMoveDown = ParameterMapping::createMapping()
                ->mapFromParent("*", "*");
        //ParameterMappingPtr mapCheckForces = ParameterMapping::createMapping()
        //        ->mapFromParent("*", "*");

        setInitState(stateMoveDown, mapMoveDown);

        //transitions
        addTransition<EvSuccess>(stateMoveDown, stateMoveDown,mapMoveDown);
        addTransition<EvFailure>(stateMoveDown, stateFailure);
        //addTransition<EvSuccess>(stateCheckForces, stateSuccess);
        //addTransition<EvFailure>(stateCheckForces, stateMoveDown);

        // has to be defined in parent states
        //addTransition<EvPlaceObjectCollision>(stateMoveDown, stateSuccess);
        // for now: success
        //addTransition<EvPlaceObjectTimeout>(stateMoveDown, stateSuccess);

        //addTransition<EvPlaceObjectTimeout>(stateCheckForces, stateSuccess);
        //addTransition<EvPlaceObjectTimeout>(stateCheckForces, stateFailure);

    }

    void StatechartPlaceObject::onEnter()
    {
        ARMARX_DEBUG << "Entering StatechartPlaceObject";

        RobotStatechartContext* context = getContext<RobotStatechartContext>();
        //setLocal("jointVelocityChannel", context->getChannelRef(context->getKinematicUnitObserverName(),"jointvelocities"));
        //ChannelRefPtr tempChannelRef = context->getChannelRef(context->getKinematicUnitObserverName(),"jointvelocities");

        std::string rnsName = getInput<std::string>("robotNodeSetName");
        std::string rnBaseName = getInput<std::string>("robotBaseFrame");
        ARMARX_LOG << "RNS:" << rnsName << ", rnBaseName:" << rnBaseName << flush;

        VirtualRobot::RobotNodeSetPtr nodeSet = context->remoteRobot->getRobotNodeSet(rnsName);
        if (!nodeSet)
        {
            throw UserException("RobotNodeSet not found: " + rnsName);
        }
        if (!nodeSet->getTCP())
        {
            throw UserException("RobotNodeSet without tcp: " + rnsName);
        }
        VirtualRobot::RobotNodePtr nodeBase = context->remoteRobot->getRobotNode(rnBaseName);
        if (!nodeBase)
        {
            throw UserException("RobotNode not found: " + rnBaseName);
        }

        Eigen::Matrix4f tcpPose = nodeSet->getTCP()->getGlobalPose();
        tcpPose = nodeBase->toLocalCoordinateSystem(tcpPose);
        FramedPosePtr fp(new FramedPose(tcpPose,rnBaseName));

        setLocal("startPose", fp);
        ARMARX_LOG << "startPose:" << tcpPose << ", rnBaseName:" << rnBaseName << flush;


        Eigen::Matrix4f tcpGoalPose = nodeSet->getTCP()->getGlobalPose();
        float maxHeight = getInput<float>("maxHeightToMoveDown");
        tcpGoalPose(2,3) -= maxHeight;
        tcpGoalPose = nodeBase->toLocalCoordinateSystem(tcpGoalPose);
        FramedPosePtr fpGoal(new FramedPose(tcpGoalPose,rnBaseName));

        setLocal("goalPose", fpGoal);
        ARMARX_LOG << "goalPose:" << tcpGoalPose << ", rnBaseName:" << rnBaseName << flush;

        ARMARX_LOG << eINFO << "Installing place object timeout condition";
        float timeoutPlaceObject = float(getInput<int>("timeoutPlaceObject"));

        condPlaceObjectTimeout = setTimeoutEvent(timeoutPlaceObject, createEvent<EvPlaceObjectTimeout>());



        // install the force condition

        float force = getInput<float>("minForceForSuccess");
        Term cond;
        for(unsigned int i=0; i<nodeSet->getSize(); i++)
        {
            ParameterList inrangeCheck;
            inrangeCheck.push_back(new Variant(-force));
            inrangeCheck.push_back(new Variant(force));

            Literal inrangeLiteral(DataFieldIdentifier(context->getKinematicUnitObserverName(),"jointtorques", nodeSet->getNode(i)->getName()), "inrange", inrangeCheck);
            cond = cond && !inrangeLiteral;
        }
        ARMARX_LOG << "Installed force consition, minForce:" << force << flush;
        condPlaceObjectCollision = installCondition<EvPlaceObjectCollision>(cond);
    }

    void StatechartPlaceObject::onExit()
    {
        removeTimeoutEvent(condPlaceObjectTimeout);
        removeCondition(condPlaceObjectCollision);

        RobotStatechartContext *context = getContext<RobotStatechartContext>();
        HandUnitInterfacePrx handUnitPrx = context->getHandUnit("RightHandUnit");
        if (handUnitPrx)
        {
            ARMARX_INFO << "Sending open hand command" << flush;
            handUnitPrx->open();
            ARMARX_INFO << "TEST: set object released: Vitalis" << flush;
            handUnitPrx->setObjectReleased("Vitalis");
        } else
            ARMARX_LOG << eINFO << "No hand unit found...";

        ARMARX_LOG << eINFO << "Exiting StatechartPlaceObject...";
    }

    // ****************************************************************
    // Implementation of StateMoveDown
    // ****************************************************************

    void StateMoveDown::defineParameters()
    {
        addToInput("maxHeightToMoveDown", VariantType::Float, false);
        addToInput("minForceForSuccess", VariantType::Float, false);
        addToInput("timeoutPlaceObject", VariantType::Int, false);
        addToInput("robotNodeSetName", VariantType::String, false);
        addToInput("robotBaseFrame", VariantType::String, false);
        addToInput("startPose", VariantType::FramedPose, false);
        addToInput("goalPose", VariantType::FramedPose, false);
    }

    void StateMoveDown::onEnter()
    {
        ARMARX_INFO << "Entering StateMoveDown" << flush;

        RobotStatechartContext* rsContext = getContext<RobotStatechartContext>();
        std::string rnsName = getInput<std::string>("robotNodeSetName");
        std::string rnBaseName = getInput<std::string>("robotBaseFrame");
        VirtualRobot::RobotPtr localRobot = RemoteRobot::createLocalClone(rsContext->robotStateComponent);

        VirtualRobot::RobotNodeSetPtr nodeSet = localRobot->getRobotNodeSet(rnsName);
        VirtualRobot::RobotNodePtr nodeBase = localRobot->getRobotNode(rnBaseName);
        FramedPosePtr startPoseFramed = getInput<FramedPose>("startPose");
        FramedPosePtr goalPoseFramed = getInput<FramedPose>("goalPose");

        Eigen::Matrix4f startPose = startPoseFramed->toEigen();
        Eigen::Matrix4f goalPose = goalPoseFramed->toEigen();
        Eigen::Matrix4f currentPose = nodeSet->getTCP()->getGlobalPose();
        currentPose = nodeBase->toLocalCoordinateSystem(currentPose);

        ARMARX_INFO << "startPose (coordsystem " << rnBaseName << ") : " << startPose << flush;
        ARMARX_INFO << "goalPose (coordsystem " << rnBaseName << ") : " << goalPose << flush;
        ARMARX_INFO << "currentPose (coordsystem " << rnBaseName << ") : " << currentPose << flush;

        Eigen::Vector3f goalVect = goalPose.block(0,3,3,1) - currentPose.block(0,3,3,1);
        float maxDist = 10.0f;
        float dist = goalVect.norm();
        if (dist<maxDist)
        {
            ARMARX_ERROR << "Did not hit any surface, sending failure event..." << flush;
            sendEvent<EvFailure>();
        }
        goalVect = goalVect.normalized() * maxDist;

        goalPose.block(0,3,3,1) = currentPose.block(0,3,3,1) + goalVect;
        ARMARX_INFO << "goalPose cut (coordsystem " << rnBaseName << ") : " << goalPose << flush;


        VirtualRobot::DifferentialIKPtr ikSolver(new VirtualRobot::DifferentialIK(nodeSet,nodeBase));

        ikSolver->setGoal(goalPose,nodeSet->getTCP(),VirtualRobot::IKSolver::Position,10.0f);

        bool ikOK = ikSolver->solveIK(0.6f,0,10);

        if (!ikOK)
        {
            ARMARX_ERROR << "No IK solution for goal pose (coordsystem " << rnBaseName << ") : " << goalPose << flush;
            sendEvent<EvFailure>();
        }

        std::vector<float> jointValues = nodeSet->getJointValues();
        NameValueMap jointNamesAndValues;
        NameControlModeMap controlModes;
        for (unsigned int j=0; j<jointValues.size(); j++)
        {
            std::string nName = nodeSet->getNode(j)->getName();
            jointNamesAndValues[nName]=jointValues[j];
            controlModes[nName] = ePositionControl;
            ARMARX_INFO << "target for joint :" << nodeSet->getNode(j)->getName() << " -> " << jointValues[j] << flush;
        }
        rsContext->kinematicUnitPrx->switchControlMode(controlModes);
        rsContext->kinematicUnitPrx->setJointAngles(jointNamesAndValues);
        sendEvent<EvSuccess>();
        ARMARX_INFO << "StateMoveDown: Done onEnter()";
    }

    void StateMoveDown::onExit()
    {
        ARMARX_INFO << "StatePlaceObject: Done onExit()";
    }


    // ****************************************************************
    // Implementation of StateCheckForces
    // ****************************************************************
/*
    void StateCheckForces::defineParameters()
    {
        addToInput("maxHeightToMoveDown", VariantType::List(VariantType::Float), false);
        addToInput("minForceForSuccess", VariantType::List(VariantType::Float), false);
        addToInput("timeoutPlaceObject", VariantType::Float, false);
        addToInput("robotNodeSetName", VariantType::String, false);
        addToInput("robotBaseFrame", VariantType::String, false);
        addToInput("startPose", VariantType::List(VariantType::FramedPose), false);
    }

    void StateCheckForces::onEnter()
    {
        ARMARX_DEBUG << "Entering StateCheckForces" << flush;

        float force = getInput<float>("minForceForSuccess");
        RobotStatechartContext* rsContext = getContext<RobotStatechartContext>();
        rsContext->kinematicUnitObserverPrx->getDataFields()


        RobotStatechartContext* rsContext = getContext<RobotStatechartContext>();
        std::string rnsName = getInput<std::string>("robotNodeSetName");
        std::string rnBaseName = getInput<std::string>("robotBaseFrame");


        VirtualRobot::RobotPtr localRobot = RemoteRobot::createLocalClone(context->robotStateComponent);
        std::string rnsName = getInput<std::string>("robotNodeSetName");
        std::string rnBaseName = getInput<std::string>("robotBaseFrame");
        VirtualRobot::RobotNodeSetPtr nodeSet = localRobot->getRobotNodeSet(rnsName);
        VirtualRobot::RobotNodePtr nodeBase = localRobot->getRobotNode(rnBaseName);
        FramedPosePtr startPoseFramed = getInput<FramedPose>("startPose");
        FramedPosePtr goalPoseFramed = getInput<FramedPose>("goalPose");

        Eigen::Matrix4f startPose = startPoseFramed->toEigen();
        Eigen::Matrix4f goalPose = goalPoseFramed->toEigen();
        Eigen::Matrix4f currentPose = nodeSet->getTCP()->getGlobalPose();
        currentPose = nodeBase->toLocalCoordinateSystem(currentPose);

        ARMARX_DEBUG << "startPose (coordsystem " << rnBaseName << ") : " << startPose << flush;
        ARMARX_DEBUG << "goalPose (coordsystem " << rnBaseName << ") : " << goalPose << flush;
        ARMARX_DEBUG << "currentPose (coordsystem " << rnBaseName << ") : " << currentPose << flush;

        float dist = (goalPose.block(0,3,3,1) - currentPose.block(0,3,3,1)).norm();
        if (dist<10.0f)
        {
            ARMARX_ERROR << "Did not hit any surface, sending failure event..." << flush;
            sendEvent<EvFailure>();
        }


        VirtualRobot::DifferentialIKPtr ikSolver(new VirtualRobot::DifferentialIKPtr(nodeSet,nodeBase));

        ikSolver->setGoal(goalPose,nodeSet->getTCP(),VirtualRobot::IKSolver::Position,10.0f);

        bool ikOK = ikSolver->solveIK(0.6f,0,10);

        if (!ikOK)
        {
            ARMARX_ERROR << "No IK solution for goal pose (coordsystem " << rnBaseName << ") : " << goalPose << flush;
            sendEvent<EvFailure>();
        }

        std::vector<float> jointValues = nodeSet->getJointValues();
        NameValueMap jointNamesAndValues;

        for (unsigned int j=0; j<jointValues->getSize(); j++)
        {
            jointNamesAndValues[nodeSet->getNode(j)->getName()]=jointValues[j];
            ARMARX_DEBUG << "target for joint :" << nodeSet->getNode(j)->getName() << " -> " << jointValues[j] << flush;
        }

        rsContext->kinematicUnitPrx->setJointAngles(jointNamesAndValues);

        ARMARX_DEBUG << "StateCheckForces: Done onEnter()";
    }

    void StateCheckForces::onExit()
    {
        ARMARX_DEBUG << "StateCheckForces: Done onExit()";
    }

*/

}

