#ifndef MOTIONCONTROL_H
#define MOTIONCONTROL_H

#include <Core/statechart/Statechart.h>
#include <RobotAPI/libraries/core/RobotStatechartContext.h>

namespace armarx
{

namespace MotionControl
{

    /**
    *  \class MotionControlHandler
    *  \ingroup MotionControl
    *  MotionControlHandler is the remote state interface for the MotionControl-
    *  States. It contains convenience states that allow to move parts of the robot,
    *  i.e. direct and inverse kinematics for arbitrary kinematic chains and predefined
    *  parts like the arms.
    */
    class MotionControlOfferer : public RemoteStateOfferer<RobotStatechartContext>
    {
    public:
        MotionControlOfferer();
        void onInitRemoteStateOfferer();
        void onConnectRemoteStateOfferer();
        std::string getStateOffererName() const { return "MotionControl"; }
    };


    DEFINEEVENT(EvJointTargetReached)
    DEFINEEVENT(EvJointVelocitiesTargetReached)
    DEFINEEVENT(EvCalculationDone)
    DEFINEEVENT(EvTimeout)


    // test states

    struct MotionControlTestState : StateTemplate<MotionControlTestState>
    {
        void defineParameters();
        void defineSubstates();
    };


    struct MotionControlTestStateIK : StateTemplate<MotionControlTestStateIK>
    {
        void defineParameters();
        void defineSubstates();
    };



    /**
    *  \ingroup MotionControl
    *   Move the joints of a kinematic chain to the desired values.
    *   \param jointNames the names of the joints to be moved
    *   \param targetJointValues the desired joint values
    *   \param targetTolerance tolerance defining how precisely the joint position has to be reached
    *   \param timeoutInMs a timeout after which the attempt to move is aborted
    */
    struct MoveJoints : StateTemplate<MoveJoints>
    {
        RobotStatechartContext* context;
        ConditionIdentifier targetReachedCondition;
        ActionEventIdentifier timeoutEvent;
        void defineParameters();
        void onEnter();
        void onBreak();
        void onExit();
    };


    /**
    *  \ingroup MotionControl
    *   Set the velocities of the joints of a kinematic chain.
    *   \param jointNames the names of the joints to be moved
    *   \param targetJointVelocities the desired joint velocities
    *   \param targetJointVelocityTolerance tolerance defining how precisely the joint velocity has to be reached
    *   \param timeoutInMs a timeout after which the attempt is aborted
    */
    struct MoveJointsVelocityControl : StateTemplate<MoveJointsVelocityControl>
    {
        RobotStatechartContext* context;
        ConditionIdentifier targetVelocitiesReachedCondition;
        ActionEventIdentifier timeoutEvent;
        void defineParameters();
        void onEnter();
        void onExit();
    };


    /**
    *  \ingroup MotionControl
    *   Move the TCP to a desired pose.
    *   \param kinematicChainName the name of the kinematic chain that is used
    *   \param targetTCPPosition the target position for the TCP
    *   \param targetTCPOrientation the target orientation
    *   \param targetPositionDistanceTolerance tolerance for the position to decide if the motion was successfull
    *   \param ikWithOrientation: Consider orientation for IK computation. if false, the rotation of the TCP is not considered.
    *   \param targetOrientationDistanceTolerance tolerance for the orientation to decide if the motion was successfull
    *   \param timeoutInMs a timeout after which the motion is aborted
    */
    struct MoveTCPPoseIK : StateTemplate<MoveTCPPoseIK>
    {
        RobotStatechartContext* context;
        ConditionIdentifier targetReachedCondition;
        ActionEventIdentifier timeoutEvent;
        void defineParameters();
        void defineSubstates();
        void onEnter();
        void onExit();
    };


    DEFINEEVENT(EvLastPointReached)
    DEFINEEVENT(EvLastPointNotYetReached)

    /**
    *  \ingroup MotionControl
    *   Move the TCP along a trajectory of poses.
    *   \param kinematicChainName the name of the kinematic chain that is used
    *   \param targetTCPPositions the list of target positions for the TCP
    *   \param targetTCPOrientations the list of target orientations
    *   \param targetPositionDistanceTolerance tolerance for the position to decide if the motion was successfull
    *   \param targetOrientationDistanceTolerance tolerance for the orientation to decide if the motion was successfull
    *   \param ikWithOrientation: Consider orientation for IK computation. if false, the rotation of the TCP is not considered.
    *   \param timeoutInMs a timeout after which the motion is aborted
    */
    struct MoveTCPTrajectory : StateTemplate<MoveTCPTrajectory>
    {
        RobotStatechartContext* context;
        ConditionIdentifier lastTargetReachedCondition;
        ActionEventIdentifier timeoutEvent;
        ActionEventIdentifier trajectoryPointCounterID;
        void defineParameters();
        void defineSubstates();
        void onEnter();
        void onExit();
    };




    /**
    *  \ingroup MotionControl
    * Stop all motion of the robot. It sets all velocities to 0. No Smooth stopping
    * is implemented here.
    */
    struct StopRobot : StateTemplate<StopRobot>
    {
        RobotStatechartContext* context;

        void onEnter();
    };


    /**
    *  \ingroup MotionControl
    * Closes the given hand of the robot
    * \param useLeftHand True if left hand should be closed, false for the right hand
    */
    struct DoPreshape : StateTemplate<DoPreshape>
    {
        void defineParameters();
        void onEnter();
        void onExit();
        static SingleTypeVariantList GetPreshapeSet(const SingleTypeVariantListPtr preshapes, const std::string& preshapePrefix);

        ConditionIdentifier cond;
        ActionEventIdentifier action;
    };

    DEFINEEVENT(EvReachedIntermediatePreshape)
    DEFINEEVENT(EvReachedFinalPreshape)
    struct SelectAndDoPreshape : StateTemplate<SelectAndDoPreshape>
    {
        void defineSubstates();
        void defineParameters();
        void onEnter();
    };

    struct DoPreshapeSet : StateTemplate<DoPreshapeSet>
    {
        void defineSubstates();
        void defineParameters();
        void onEnter();
        void onExit();

    };





    struct DoPrefixPreshapeSet : StateTemplate<DoPrefixPreshapeSet>
    {
        void defineSubstates();
        void defineParameters();
        void onEnter();
    };





    /**
    *  \ingroup MotionControl
    * Opens the given hand of the robot
    * \param useLeftHand True if left hand should be opened, false for the right hand
    */
    struct OpenHand : StateTemplate<OpenHand>
    {
        void defineSubstates();
        void defineParameters();
        void onEnter();
    };


    /**
    *  \ingroup MotionControl
    * Closes the given hand of the robot
    * \param useLeftHand True if left hand should be closed, false for the right hand
    */
    struct CloseHand : StateTemplate<CloseHand>
    {
        void defineSubstates();
        void defineParameters();
        void onEnter();
    };




    /**
    *  \ingroup MotionControl
    * Moves the head so that it looks at the specified target position
    * \param target Position at which the robot will try to look
    * \param headIKKinematicChainName Name of the kinematic chain for the head that will be sued in the IK
    */
    struct HeadLookAtTarget : StateTemplate<HeadLookAtTarget>
    {
        void defineParameters();
        void defineSubstates();
    };




    /////////////////////////////////////////////////////////////////////
    //// intermediate states
    /////////////////////////////////////////////////////////////////////

    struct CalculateJointAngleConfiguration : StateTemplate<CalculateJointAngleConfiguration>
    {

        ActionEventIdentifier timeoutEvent;
        void defineParameters();
        void run();

    };

    struct ValidateJointAngleConfiguration : StateTemplate<ValidateJointAngleConfiguration>
    {
        RobotStatechartContext* context;
        ActionEventIdentifier timeoutEvent;
        void defineParameters();
        void onEnter();
    };




    struct MoveTCPTrajectoryInit : StateTemplate<MoveTCPTrajectoryInit>
    {
        void defineParameters();
        void onEnter();
    };




    struct MoveTCPTrajectoryCheckCounter : StateTemplate<MoveTCPTrajectoryCheckCounter>
    {
        RobotStatechartContext* context;
        void defineParameters();
        void onEnter();
    };




    struct MoveTCPTrajectoryNextPoint : StateTemplate<MoveTCPTrajectoryNextPoint>
    {
        RobotStatechartContext* context;
        void defineParameters();
        void defineSubstates();
        void onEnter();
    };




    struct CalculateHeadIK : StateTemplate<CalculateHeadIK>
    {
        void defineParameters();
        void onEnter();
        void run();
        FramedPositionPtr targetPosition;
        std::string kinematicChainName;
        std::string virtualPrismaticJointName;
    };

}

}

#endif
