/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2 of
* the License, or (at your option) any later version.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU Lesser General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @package    ArmarX::
* @author     Mirko Waechter ( mirko.waechter at kit dot edu)
* @date       2013
* @copyright  http://www.gnu.org/licenses/gpl.txt
*             GNU General Public License
*/

#include "ZeroForceControl.h"

#include <RobotAPI/libraries/robotstate/remote/RobotStateObjectFactories.h>
#include <Core/core/exceptions/local/ExpressionException.h>

#include <RobotAPI/libraries/core/RobotStatechartContext.h>

namespace armarx {

    ZeroForceControl::ZeroForceControl()
    {
    }



    void ZeroForceControl::defineSubstates()
    {
        StatePtr calc = addState<ZeroForceControlForceToAcc>("ZeroForceControlForceToAcc");

        PMPtr mapping = PM::createMapping()
                ->mapFromParent("*")
                ->mapFromDataField(new DataFieldIdentifier(".."), "currentForce")
                ->mapFromDataField(new DataFieldIdentifier(".."), "currentTorque")
                ;

        setInitState(calc, mapping);
        addTransition<EvSensorUpdate>(calc,calc, mapping->mapFromOutput("*") );
    }

    void ZeroForceControl::defineParameters()
    {
        addToInput("robotNodeSetName",VariantType::String, false);
        addToInput("tcpName",VariantType::String, false);
        addToInput("sensitivity",VariantType::Float, false);
        addToInput("maxAcc",VariantType::Float, false);

        addToLocal("currentSensitivity",VariantType::Float);
        addToLocal("currentVelocity",VariantType::Float);
        addToLocal("currentAcc",VariantType::Float);
        addToLocal("timestamp",VariantType::Float);
    }

    void ZeroForceControl::onEnter()
    {
        RobotStatechartContext* c = getContext<RobotStatechartContext>();
        c->tcpControlPrx->request();
        Eigen::Vector3f null(3);
        null.setZero();
        setLocal("currentSensitivity", 0.0f);
        setLocal("currentVelocity", new FramedVector3(null, getInput<std::string>("tcpName")));
        setLocal("currentAcc", new FramedVector3(null, getInput<std::string>("tcpName")));
        setLocal("timestamp", (float)IceUtil::Time::now().toMilliSecondsDouble());
    }

    void ZeroForceControl::onExit()
    {
        RobotStatechartContext* c = getContext<RobotStatechartContext>();
        c->tcpControlPrx->release();
    }


    ////////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////

    ZeroForceControlForceToAcc::ZeroForceControlForceToAcc()
    {

    }

    void ZeroForceControlForceToAcc::defineParameters()
    {
        addToInput("robotNodeSetName",VariantType::String, false);
        addToInput("sensitivity",VariantType::Float, false);
        addToInput("maxAcc",VariantType::Float, false);


        addToInput("currentSensitivity",VariantType::Float, false);
        addToInput("currentVelocity",VariantType::FramedVector3, false);
        addToInput("currentAcc",VariantType::FramedVector3, false);
        addToInput("currentForce",VariantType::FramedVector3, false);
        addToInput("currentTorque",VariantType::FramedVector3, false);
        addToInput("timestamp",VariantType::Float, false);

        addToOutput("currentSensitivity",VariantType::Float, false);
        addToOutput("currentVelocity",VariantType::Float, false);
        addToOutput("currentAcc",VariantType::Float, false);

    }

    void ZeroForceControlForceToAcc::onEnter()
    {
        Literal update("ForceTorqueUnit","updated", Literal::createParameterList());
        installCondition<EvSensorUpdate>(update);
        IceUtil::Time duration = IceUtil::Time::now() - IceUtil::Time::milliSecondsDouble(getInput<float>("timestamp"));

        std::string robotNodeSetName = getInput<std::string>("robotNodeSetName");
//        FramedVector3Ptr vel = getInput<FramedVector3>("currentVelocity");
        FramedVector3Ptr vel = getInput<FramedVector3>("currentVelocity");
        FramedVector3Ptr curAcc = getInput<FramedVector3>("currentAcc");
        FramedVector3Ptr curForce = getInput<FramedVector3>("currentForce");
        ARMARX_CHECK_EXPRESSION(vel->frame == curAcc->frame);
        ARMARX_CHECK_EXPRESSION(vel->frame == curForce->frame);
        float forceThreshold = 2.0f;
        //float maxSensitivity = 2.0f;

        Eigen::Vector3f newVel(3);
        Eigen::Vector3f newAcc(3);
        if(curForce->toEigen().norm() > forceThreshold){
            newAcc = 2 *  curForce->toEigen() * 5;
        }
        else
        {
            newAcc = -10*vel->toEigen().normalized();
        }

        newVel = vel->toEigen() + newAcc * duration.toMilliSecondsDouble()*0.001;


        setOutput("currentAcc", Variant(new FramedVector3(newAcc, curAcc->frame)));
        RobotStatechartContext* c = getContext<RobotStatechartContext>();

        c->tcpControlPrx->setTCPVelocity(robotNodeSetName, "", new FramedVector3(newVel, vel->frame), NULL);
    }

    void ZeroForceControlForceToAcc::onExit()
    {

    }

}
